import uuid

import numpy as np, weakref
from ..JHTML import JHTML, DefaultOutputArea, JupyterAPIs

__all__ = [
    "Var",
    "DefaultVars",
    "InterfaceVars",
    "VariableSynchronizer",
    "VariableNamespace",
    "WidgetControl"
]

__reload_hook__ = ['..JHTML']

class SettingChecker:
    int_types = (int, np.integer)
    numerics_types = (int, np.integer, float, np.floating)
    control_type = None
    @classmethod
    def check(self, **props):
        return False
    checkers = []
class CheckboxChecker(SettingChecker):
    control_type = "Checkbox"
    @classmethod
    def check(self, value=None, **rest):
        return isinstance(value, bool)
SettingChecker.checkers.append(CheckboxChecker)
class DropdownChecker(SettingChecker):
    control_type = "Dropdown"
    @classmethod
    def check(self, options=None, **rest):
        return options is not None
SettingChecker.checkers.append(DropdownChecker)
class TextChecker(SettingChecker):
    control_type = "Text"
    @classmethod
    def check(self, value=None, **rest):
        return isinstance(value, str)
SettingChecker.checkers.append(TextChecker)
class IntSliderChecker(SettingChecker):
    control_type = "IntSlider"
    @classmethod
    def check(self, value=None, min=None, max=None, **rest):
        return isinstance(value, self.int_types) and isinstance(min, self.int_types) and isinstance(max, self.int_types)
SettingChecker.checkers.append(IntSliderChecker)
class FloatSliderChecker(SettingChecker):
    control_type = "FloatSlider"
    @classmethod
    def check(self, value=None, min=None, max=None, **rest):
        return (
                isinstance(value, self.numerics_types)
                and isinstance(min, self.numerics_types)
                and isinstance(max, self.numerics_types)
        )
SettingChecker.checkers.append(FloatSliderChecker)
class IntRangeChecker(SettingChecker):
    control_type = "IntRangeSlider"
    @classmethod
    def check(self, value=None, min=None, max=None, **rest):
        return (
                len(value) == 2
                and isinstance(value[0], self.int_types)
                and isinstance(value[1], self.int_types)
                and isinstance(min, self.int_types) and isinstance(max, self.int_types)
        )
SettingChecker.checkers.append(IntRangeChecker)
class FloatRangeChecker(SettingChecker):
    control_type = "FloatRangeSlider"
    @classmethod
    def check(self, value=None, min=None, max=None, **rest):
        return (
                len(value) == 2
                and isinstance(value[0], self.numerics_types)
                and isinstance(value[1], self.numerics_types)
                and isinstance(min, self.numerics_types) and isinstance(max, self.numerics_types)
        )
SettingChecker.checkers.append(FloatRangeChecker)

class InterfaceVars:
    _cache_stack = []
    def __init__(self, *vars, callbacks=None):
        vars = [Var(x) for x in vars]
        self._var_set = set(vars)
        self.var_list = vars
        self.callbacks = set(callbacks) if callbacks is not None else set()
    @classmethod
    def active_vars(cls):
        if len(cls._cache_stack) > 0:
            return cls._cache_stack[-1]
        else:
            return None
    @property
    def dict(self):
        return {v.name:v.value for v in self.var_list}
    @property
    def items(self):
        return [(v.name, v.value) for v in self.var_list]
    def __iter__(self):
        return iter(self.var_list)
    def add(self, var):
        if var not in self._var_set:
            self.var_list.append(var)
            self._var_set.add(var)
            for c in self.callbacks:
                c(self)
    def __enter__(self):
        self._cache_stack.append(self)
        return self.var_list
    def __exit__(self, exc_type, exc_val, exc_tb):
        self._cache_stack.pop()
    def __repr__(self):
        return "{}({})".format(type(self).__name__, ", ".join(s.name for s in self._var_set))
class DefaultVars:
    _var_stack = []
    def __init__(self, vars:InterfaceVars=None):
        self.vars = InterfaceVars() if vars is None else vars
    def __enter__(self):
        self._var_stack.append(self)
        self.vars.__enter__()
        return self.vars
    def __exit__(self, exc_type, exc_val, exc_tb):
        self._var_stack.pop()
        self.vars.__exit__(exc_type, exc_val, exc_tb)
    @classmethod
    def resolve(cls):
        return InterfaceVars() if len(cls._var_stack) == 0 else cls._var_stack[-1].vars
class VariableNamespace:
    _namespace_cache = weakref.WeakValueDictionary()
    def __init__(self, name=None, dedupe=True):
        if name is None:
            name = uuid.uuid4()
        self.name = name
        self._old_space = None
        if dedupe:
            if name in self._namespace_cache:
                self._var_cache = self._namespace_cache[name]._var_cache
            else:
                self._var_cache = weakref.WeakValueDictionary()
                self._namespace_cache[name] = self
        else:
            self._var_cache = weakref.WeakValueDictionary()
    def __repr__(self):
        return "{}({}, {})".format(
            type(self).__name__,
            self.name,
            list(self._var_cache.values())
        )
    @classmethod
    def create(cls, name):
        if isinstance(name, VariableNamespace):
            return name
        else:
            if name not in cls._namespace_cache:
                this = VariableNamespace(name)
                cls._namespace_cache[name] = this # hold a reference...
            return cls._namespace_cache[name]
    def __contains__(self, item):
        return self._var_cache.__contains__(item)
    def __getitem__(self, item):
        return self._var_cache[item]
    def __setitem__(self, key, value):
        self._var_cache[key] = value
    def __enter__(self):
        self._old_space = VariableSynchronizer.current_namespace
        VariableSynchronizer.current_namespace = self
    def __exit__(self, exc_type, exc_val, exc_tb):
        VariableSynchronizer.current_namespace = self._old_space
        self._old_space = None
class VariableSynchronizer:
    current_namespace = VariableNamespace.create('globals')
    def __init__(self, name, namespace=None, value=None, callbacks=(), output_pane=None):
        self._name = name
        self.namespace = namespace
        self._value = value
        self.callbacks = weakref.WeakSet(callbacks)
        self.output_pane = DefaultOutputArea.get_default() if output_pane is None else output_pane
        self._watchers = weakref.WeakSet()
    def __repr__(self):
        return "{}({}, {!r})".format(
            type(self).__name__,
            self._name,
            self._value
        )
    @classmethod
    def create_var(cls, var):
        var_cache = InterfaceVars.active_vars()
        if isinstance(var, VariableSynchronizer):
            if var_cache is not None:
                var_cache.add(var)
            return var
        else:
            if var not in cls.current_namespace:
                this_var = VariableSynchronizer(var, namespace=cls.current_namespace)
                cls.current_namespace[var] = this_var
            if var_cache is not None:
                var_cache.add(cls.current_namespace[var])
            return cls.current_namespace[var]
    @property
    def name(self):
        return self._name
    @property
    def value(self):
        return self._value
    @value.setter
    def value(self, v):
        self.set_value(v)
    def set_value(self, v, caller=None):
        with self.output_pane:
            old = self._value
            try:
                check = old != v
            except TypeError:
                check = old is not v
            if check:
                self._value = v
                for c in self.callbacks:
                    c({'var': self, 'old': old, 'new': self._value})
                for w in self._watchers:
                    if w is not caller:
                        w.value = self._value
    def link(self, widget):
        if hasattr(widget, 'value'):
            self.set_value(widget.value, caller=widget)
            widget.observe(lambda d: self.set_value(widget.value, caller=widget), names=['value'])
        self._watchers.add(widget)
def Var(name, namespace=None):
    if namespace is not None:
        with VariableNamespace.create(namespace):
            return VariableSynchronizer.create_var(name)
    else:
        return VariableSynchronizer.create_var(name)
class WidgetControl:
    def __init__(self, var, control_type=None, widget=None, **settings):
        self.var = VariableSynchronizer.create_var(var)
        self.settings = settings
        if widget is None:
            widget = self._build_widget(control_type, settings)
        self.widget = widget
    @classmethod
    def _build_widget(cls, control_type, settings):
        if control_type is None:
            for checker in SettingChecker.checkers:
                if checker.check(**settings):
                    control_type = getattr(JupyterAPIs.get_widgets_api(), checker.control_type)
                    break
            else:
                control_type = JHTML.OutputArea
        return control_type(**settings)

    def to_widget(self):
        self.var.link(self.widget)
        return self.widget


"""
Convenience functions that are inefficient, but are maybe a bit easier to work with?
"""

from .CoordinateSystems import *
from collections import namedtuple
import numpy as np

__all__ = [
    "cartesian_to_zmatrix",
    "zmatrix_to_cartesian"
]

def cartesian_to_zmatrix(coords, ordering=None, use_rad = True):
    """
    Converts Cartesians to Z-Matrix coords and returns the underlying arrays

    :param coords: input Cartesians
    :type coords: np.ndarray
    :param ordering: the Z-matrix ordering as a list of lists
    :type ordering:
    :return: Z-matrix coords
    :rtype: Iterable[np.ndarray | list]
    """

    zms = CoordinateSet(coords, system=CartesianCoordinates3D).convert(ZMatrixCoordinates, ordering=ordering, use_rad=use_rad)
    crds = namedtuple("zms", ["coords", "ordering"])
    return crds(np.asarray(zms), zms.converter_options['ordering'])

def zmatrix_to_cartesian(coords, ordering=None, origins=None, axes=None, use_rad=True):
    """
    Converts Z-maztrix coords to Cartesians

    :param coords:
    :type coords: np.ndarray
    :param ordering:
    :type ordering:
    :param origins:
    :type origins:
    :param axes:
    :type axes:
    :param use_rad:
    :type use_rad:
    :return:
    :rtype:
    """
    carts = CoordinateSet(coords, ZMatrixCoordinates).convert(CartesianCoordinates3D,
                                                            ordering=ordering, origins=origins, axes=axes, use_rad=use_rad)
    return np.asarray(carts)


import re, os

__all__ = [
    "StringMatcher",
    "MatchList",
    "FileMatcher"
]

class StringMatcher:
    """
    Defines a simple filter that applies to a file and determines whether or not it matches the pattern
    """

    def __init__(self, match_patterns, negative_match = False):
        if isinstance(match_patterns, str):
            pattern = re.compile(match_patterns)
            self.matcher = lambda f, p = pattern: re.match(p, f)
        elif hasattr(re, "Pattern") and isinstance(match_patterns, re.Pattern): # re.Pattern is new as of python3.7...
            self.matcher = lambda f, p = match_patterns: re.match(p, f)
        elif type(match_patterns).__name__=="SRE_Pattern": # pre 3.7
            self.matcher = lambda f, p = match_patterns: re.match(p, f)
        elif isinstance(match_patterns, StringMatcher):
            self.matcher = match_patterns.matches
        elif callable(match_patterns):
            self.matcher = match_patterns
        else:
            ff = type(self)
            match_patterns = tuple(ff(m) if not isinstance(m, StringMatcher) else m for m in match_patterns)
            self.matcher = lambda f, p = match_patterns: all(m.matches(f) for m in p)

        self.negate = negative_match

    def matches(self, f):
        m = self.matcher(f)
        if self.negate:
            m = not m
        return m

class MatchList(StringMatcher):
    """
    Defines a set of matches that must be matched directly (uses `set` to make this basically a constant time check)
    """

    def __init__(self, *matches, negative_match = False):
        self.match_list = set(matches)
        super().__init__(lambda f, m=self.test_match: m(f), negative_match = negative_match)
    def test_match(self, f):
        return f in self.match_list

class FileMatcher(StringMatcher):
    """
    Defines a filter that uses StringMatcher to specifically match files
    """

    def __init__(self, match_patterns, negative_match = False, use_basename = False):
        super().__init__(match_patterns, negative_match = negative_match)
        self.use_basename = use_basename

    def matches(self, f):
        f_name = f if not self.use_basename else os.path.basename(f)
        return super().matches(f_name)


import functools

__all__ = [
    "mixedmethod"
]

class mixedmethod:
    def __init__(self, wrapped_fn):
        self.wrapped_fn = wrapped_fn
    def __get__(self, obj, obj_type=None):
        if obj_type is None:
            obj_type = type(obj)
        return functools.partial(self.wrapped_fn, obj if obj is not None else obj_type)


"""
Provides a set of singleton objects that can declare their purpose a little bit better than None can
"""

import enum

__all__ = [ "Default" ]

class DefaultType:
    """
    A type for declaring an argument should use its default value (for when `None` has meaning)
    """
Default=DefaultType()


"""
A package for managing extension modules.
The existing `ExtensionLoader` will be moving here, and will be supplemented by classes for dealing with compiled extensions
"""




__all__ = []
from .CLoader import *; from .CLoader import __all__ as exposed
__all__ += exposed
from .ModuleLoader import *; from .ModuleLoader import __all__ as exposed
__all__ += exposed
from .ArgumentSignature import *; from .ArgumentSignature import __all__ as exposed
__all__ += exposed
from .SharedLibraryManager import *; from .SharedLibraryManager import __all__ as exposed
__all__ += exposed
from .FFI import *; from .FFI import __all__ as exposed
__all__ += exposed


"""
A set of concrete parser objects for general use
"""

from .StringParser import *
from .RegexPatterns import *

__all__= [
    "XYZParser"
]

XYZParser = StringParser(
    RegexPattern(
        (
            Named(PositiveInteger, "NumberOfAtoms"),
            Named(Repeating(Any, min = None), "Comment", dtype=str),
            Named(
                Repeating(
                    RegexPattern(
                        (
                            Capturing(AtomName),
                            Capturing(
                                Repeating(Capturing(Number), min = 3, max = 3, suffix=Optional(Whitespace)),
                                handler= StringParser.array_handler(shape = (None, 3))
                                )
                        ),
                        joiner=Whitespace
                    ),
                    suffix=Optional(Newline)
                ),
                "Atoms"
            )
        ),
        "XYZ",
        joiner=Newline
    )
)


# there's a subtle difference between Duplicated and Repeating
# Duplicated copies the pattern directly a set number of times which allows it to
# capture every single instance of the pattern
# Repeating uses Regex syntax to repeat a pattern a potentially unspecified number of times
# which means the parser will only return the first case when asked for the groups


from .HTML import HTML
from .HTMLWidgets import ActiveHTMLWrapper, HTMLWidgets
from .Bootstrap import Bootstrap3, Bootstrap4, Bootstrap5

__all__ = ["BootstrapWidgets"]
__reload_hook__ = [".HTML", ".HTMLWidgets", ".Bootstrap"]

class BootstrapWidgetsBase:
    cdn_loader = None
    cdn_css = []
    cdn_js = []
    _cdn_cache = {}
    bootstrap_version = None

    @classmethod
    def load(cls):
        """
        Embeds Bootstrap style definitions into the active notebook

        :return:
        :rtype:
        """
        from IPython.core.display import HTML as IPyHTML
        from urllib.request import urlopen

        if cls.cdn_loader is not None:
            loader = cls.cdn_loader
        else:
            loader = ""
            for sheet in cls.cdn_css:
                if sheet in cls._cdn_cache:
                    styles = cls._cdn_cache[sheet]
                else:
                    styles = urlopen(sheet).read().decode()
                    cls._cdn_cache[sheet] = styles
                loader += "\n<style>"+styles+"</style>"
            for package in cls.cdn_js:
                if package in cls._cdn_cache:
                    js = cls._cdn_cache[package]
                else:
                    js = urlopen(package).read().decode()
                    cls._cdn_cache[package] = js
                loader += "\n<script>"+js+"</style>"
        loader_HTML = loader + """
        <div class="alert alert-info">Bootstrap loaded <p class="d-none" style="color:red">This text will be visible if Bootstrap isn't loaded</p></div>
        """

        # print(loader_HTML)
        return IPyHTML(loader_HTML)

    @classmethod
    def _monkey_patch(cls):
        for key, value in vars(cls.bootstrap_version).items():
            if (
                    not hasattr(cls, key)
                    and isinstance(value, type)
                    and issubclass(value, HTML.XMLElement)
            ): setattr(cls, key, type(key, (ActiveHTMLWrapper,), dict(base=value)) )

    @classmethod
    def Grid(cls, rows, row_attributes=None, item_attributes=None, auto_size=True, **attrs):
        return ActiveHTMLWrapper(rows, base=cls.bootstrap_version.Grid, row_attributes=row_attributes,
                                       item_attributes=item_attributes, auto_size=auto_size, **attrs)

class Bootstrap3Widgets(BootstrapWidgetsBase):
    cdn_loader = """
    <!-- Latest compiled and minified CSS -->
<link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/bootstrap@3.3.7/dist/css/bootstrap.min.css" integrity="sha384-BVYiiSIFeK1dGmJRAkycuHAHRg32OmUcww7on3RYdg4Va+PmSTsz/K68vbdEjh4u" crossorigin="anonymous">
<!-- Optional theme -->
<link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/bootstrap@3.3.7/dist/css/bootstrap-theme.min.css" integrity="sha384-rHyoN1iRsVXV4nD0JutlnGaslCJuC7uwjduW9SVrLvRYooPp2bWYgmgJQIXwl/Sp" crossorigin="anonymous">
<!-- Latest compiled and minified JavaScript -->
<script src="https://cdn.jsdelivr.net/npm/bootstrap@3.3.7/dist/js/bootstrap.min.js" integrity="sha384-Tc5IQib027qvyjSMfHjOMaLkfuWVxZxUPnCJA7l2mCWNIpG9mGCD8wGNIcPD7Txa" crossorigin="anonymous"></script>
"""
    bootstrap_version = Bootstrap3
    class Icon(ActiveHTMLWrapper): base = Bootstrap3.Icon
    class Alert(ActiveHTMLWrapper): base = Bootstrap3.Alert
    class Badge(ActiveHTMLWrapper): base = Bootstrap3.Badge
    class PanelBody(ActiveHTMLWrapper): base = Bootstrap3.PanelBody
    class PanelHeader(ActiveHTMLWrapper): base = Bootstrap3.PanelHeader
    class Panel(ActiveHTMLWrapper): base = Bootstrap3.Panel
    class Jumbotron(ActiveHTMLWrapper): base = Bootstrap3.Jumbotron
    class Col(ActiveHTMLWrapper): base = Bootstrap3.Col
    class Row(ActiveHTMLWrapper): base = Bootstrap3.Row
    class Container(ActiveHTMLWrapper): base = Bootstrap3.Container
    class Button(ActiveHTMLWrapper): base = Bootstrap3.Button
    class LinkButton(HTML.Anchor): base = Bootstrap3.LinkButton
    class Table(ActiveHTMLWrapper): base = Bootstrap3.Table
    class ListGroup(ActiveHTMLWrapper): base = Bootstrap3.ListGroup
    class ListGroupItem(ActiveHTMLWrapper): base = Bootstrap3.ListGroupItem
    class FontAwesomeIcon(ActiveHTMLWrapper): base = Bootstrap3.FontAwesomeIcon
    class GlyphIcon(ActiveHTMLWrapper): base = Bootstrap3.GlyphIcon
    class Label(ActiveHTMLWrapper): base = Bootstrap3.Label
    class ListComponent(ActiveHTMLWrapper): base = Bootstrap3.ListComponent
    class ListItemComponent(ActiveHTMLWrapper): base = Bootstrap3.ListItemComponent
    class Breadcrumb(ActiveHTMLWrapper): base = Bootstrap3.Breadcrumb
    class BreadcrumbItem(ActiveHTMLWrapper): base = Bootstrap3.BreadcrumbItem
Bootstrap3Widgets._monkey_patch()

class Bootstrap4Widgets(BootstrapWidgetsBase):
    cdn_loader = """
            <link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/bootstrap@4.6.1/dist/css/bootstrap.min.css" integrity="sha384-zCbKRCUGaJDkqS1kPbPd7TveP5iyJE0EjAuZQTgFLD2ylzuqKfdKlfG/eSrtxUkn" crossorigin="anonymous">
            <script src="https://cdn.jsdelivr.net/npm/jquery@3.5.1/dist/jquery.slim.min.js" integrity="sha384-DfXdz2htPH0lsSSs5nCTpuj/zy4C+OGpamoFVy38MVBnE+IbbVYUew+OrCXaRkfj" crossorigin="anonymous"></script>
            <script src="https://cdn.jsdelivr.net/npm/bootstrap@4.6.1/dist/js/bootstrap.bundle.min.js" integrity="sha384-fQybjgWLrvvRgtW6bFlB7jaZrFsaBXjsOMm/tB9LTS58ONXgqbR9W8oWht/amnpF" crossorigin="anonymous"></script>
            """
    bootstrap_version = Bootstrap4
    class Icon(ActiveHTMLWrapper): base = Bootstrap4.Icon
    class Alert(ActiveHTMLWrapper): base = Bootstrap4.Alert
    class Badge(ActiveHTMLWrapper): base = Bootstrap4.Badge
    class CardBody(ActiveHTMLWrapper): base = Bootstrap4.CardBody
    class CardHeader(ActiveHTMLWrapper): base = Bootstrap4.CardHeader
    class CardFooter(ActiveHTMLWrapper): base = Bootstrap4.CardFooter
    class CardImage(ActiveHTMLWrapper): base = Bootstrap4.CardImage
    class Card(ActiveHTMLWrapper): base = Bootstrap4.Card
    class Jumbotron(ActiveHTMLWrapper): base = Bootstrap4.Jumbotron
    class Col(ActiveHTMLWrapper): base = Bootstrap4.Col
    class Row(ActiveHTMLWrapper): base = Bootstrap4.Row
    class Container(ActiveHTMLWrapper): base = Bootstrap4.Container
    class Button(ActiveHTMLWrapper): base = Bootstrap4.Button
    class LinkButton(HTML.Anchor): base = Bootstrap4.LinkButton
    class Table(ActiveHTMLWrapper): base = Bootstrap4.Table
    class ListGroup(ActiveHTMLWrapper): base = Bootstrap4.ListGroup
    class ListGroupItem(ActiveHTMLWrapper): base = Bootstrap4.ListGroupItem
    class FontAwesomeIcon(ActiveHTMLWrapper): base = Bootstrap4.FontAwesomeIcon
    class GlyphIcon(ActiveHTMLWrapper): base = Bootstrap4.GlyphIcon
    class Label(ActiveHTMLWrapper): base = Bootstrap4.Label
    class Pill(ActiveHTMLWrapper): base = Bootstrap4.Pill
    class ListComponent(ActiveHTMLWrapper): base = Bootstrap4.ListComponent
    class ListItemComponent(ActiveHTMLWrapper): base = Bootstrap4.ListItemComponent
    class Breadcrumb(ActiveHTMLWrapper): base = Bootstrap4.Breadcrumb
    class BreadcrumbItem(ActiveHTMLWrapper): base = Bootstrap4.BreadcrumbItem
Bootstrap4Widgets._monkey_patch()

class Bootstrap5Widgets(BootstrapWidgetsBase):
    cdn_loader = """
            <!-- CSS only -->
            <link href="https://cdn.jsdelivr.net/npm/bootstrap@5.1.3/dist/css/bootstrap.min.css" rel="stylesheet" integrity="sha384-1BmE4kWBq78iYhFldvKuhfTAU6auU8tT94WrHftjDbrCEXSU1oBoqyl2QvZ6jIW3" crossorigin="anonymous">
            <!-- JavaScript Bundle with Popper -->
            <script src="https://cdn.jsdelivr.net/npm/bootstrap@5.1.3/dist/js/bootstrap.bundle.min.js" integrity="sha384-ka7Sk0Gln4gmtz2MlQnikT1wXgYsOg+OMhuP+IlRH9sENBO0LRn5q+8nbTov4+1p" crossorigin="anonymous"></script>
            <link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/bootstrap-icons@1.8.1/font/bootstrap-icons.css">
            """
    cdn_css = [
        "https://cdn.jsdelivr.net/npm/bootstrap@5.1.3/dist/css/bootstrap.min.css",
        "https://cdn.jsdelivr.net/npm/bootstrap-icons@1.8.1/font/bootstrap-icons.css"
    ]
    cdn_js = [
        "https://cdn.jsdelivr.net/npm/bootstrap@5.1.3/dist/js/bootstrap.bundle.min.js"
    ]
    bootstrap_version = Bootstrap5
    class Icon(HTMLWidgets.WrappedHTMLElement): base = Bootstrap5.Icon
    class Alert(HTMLWidgets.WrappedHTMLElement): base = Bootstrap5.Alert
    class Badge(HTMLWidgets.WrappedHTMLElement): base = Bootstrap5.Badge
    class CardBody(HTMLWidgets.WrappedHTMLElement): base = Bootstrap5.CardBody
    class CardHeader(HTMLWidgets.WrappedHTMLElement): base = Bootstrap5.CardHeader
    class CardFooter(HTMLWidgets.WrappedHTMLElement): base = Bootstrap5.CardFooter
    class CardImage(HTMLWidgets.WrappedHTMLElement): base = Bootstrap5.CardImage
    class Card(HTMLWidgets.WrappedHTMLElement): base = Bootstrap5.Card
    class Col(HTMLWidgets.WrappedHTMLElement): base = Bootstrap5.Col
    class Row(HTMLWidgets.WrappedHTMLElement): base = Bootstrap5.Row
    class Container(HTMLWidgets.WrappedHTMLElement): base = Bootstrap5.Container
    class ButtonGroup(HTMLWidgets.WrappedHTMLElement): base = Bootstrap5.ButtonGroup
    class Button(HTMLWidgets.WrappedHTMLElement): base = Bootstrap5.Button
    class CloseButton(HTMLWidgets.WrappedHTMLElement): base = Bootstrap5.CloseButton
    class LinkButton(HTMLWidgets.WrappedHTMLElement): base = Bootstrap5.LinkButton
    class Table(HTMLWidgets.WrappedHTMLElement): base = Bootstrap5.Table
    class ListGroup(HTMLWidgets.WrappedHTMLElement): base = Bootstrap5.ListGroup
    class ListGroupItem(HTMLWidgets.WrappedHTMLElement): base = Bootstrap5.ListGroupItem
    class FontAwesomeIcon(HTMLWidgets.WrappedHTMLElement): base = Bootstrap5.FontAwesomeIcon
    class GlyphIcon(HTMLWidgets.WrappedHTMLElement): base = Bootstrap5.GlyphIcon
    class Label(HTMLWidgets.WrappedHTMLElement): base = Bootstrap5.Label
    class Pill(HTMLWidgets.WrappedHTMLElement): base = Bootstrap5.Pill
    class ListComponent(HTMLWidgets.WrappedHTMLElement): base = Bootstrap5.ListComponent
    class ListItemComponent(HTMLWidgets.WrappedHTMLElement): base = Bootstrap5.ListItemComponent
    class Breadcrumb(HTMLWidgets.WrappedHTMLElement): base = Bootstrap5.Breadcrumb
    class BreadcrumbItem(HTMLWidgets.WrappedHTMLElement): base = Bootstrap5.BreadcrumbItem
Bootstrap5Widgets._monkey_patch()

BootstrapWidgets = Bootstrap5Widgets
# JupyterHTMLWrapper._widget_sources.append(BootstrapWidgets)


"""Defines components of an .fchk file that are already known and parseable"""

from ..Parsers.RegexPatterns import *
from .FChkDerivatives import *
import numpy as np

########################################################################################################################
#
#                                           FormattedCheckpointComponents
#
#region FormattedCheckpointComponents
FormattedCheckpointComponents = { } # we'll register on this bit by bit
# each registration should look like:

# FormattedCheckpointComponents["Name"] = parser

#endregion

########################################################################################################################
#
#                                          Int Atom Types
#

#region IInt Atom Types

def get_names(atom_ints, reader=None):
    from ..Data import AtomData
    return [ AtomData[x, "Symbol"] for x in atom_ints ]
FormattedCheckpointComponents["Int Atom Types"] = get_names

#endregion

########################################################################################################################
#
#                                          Current cartesian coordinates
#

#region Current cartesian coordinates

def reformat(coords, reader=None):
    import numpy as np

    ncoords = len(coords)
    return np.reshape(coords, (int(ncoords/3), 3))
FormattedCheckpointComponents["Current cartesian coordinates"] = reformat

#endregion

########################################################################################################################
#
#                                           Cartesian Force Constants
#

#region Cartesian Force Constants

FormattedCheckpointComponents["Cartesian Force Constants"] = FchkForceConstants

#endregion

########################################################################################################################
#
#                                           Cartesian 3rd/4th derivatives
#

#region Cartesian 3rd/4th derivatives

FormattedCheckpointComponents["Cartesian 3rd/4th derivatives"] = FchkForceDerivatives

#endregion

########################################################################################################################
#
#                                           Dipole Derivatives
#

#region Dipole Derivatives

FormattedCheckpointComponents["Dipole Derivatives"] = FchkDipoleDerivatives

#endregion

########################################################################################################################
#
#                                           Dipole Derivatives num derivs
#

#region Dipole Derivatives num derivs

FormattedCheckpointComponents["Dipole Moment num derivs"] = FchkDipoleNumDerivatives

#region Dipole Derivatives num derivs

FormattedCheckpointComponents["Dipole Derivatives num derivs"] = FchkDipoleHigherDerivatives

#endregion

########################################################################################################################
#
#                                           Vib-Modes
#

#region Vib-Modes

def split_vib_modes(mcoeffs, reader=None):
    """Pulls the mode vectors from the coeffs
    There should be 3N-6 modes where each vector is 3N long so N = (1 + sqrt(1 + l/9))

    :param mcoeffs:
    :type mcoeffs:
    :return:
    :rtype:
    """
    import numpy as np

    # l = len(mcoeffs)
    n = reader.num_atoms#int(1 + np.sqrt(1 + l/9))
    return np.reshape(mcoeffs, (-1, 3*n))
FormattedCheckpointComponents["Vib-Modes"] = split_vib_modes

#endregion

########################################################################################################################
#
#                                           Vib-E2
#

#region Vib-E2

def split_vib_e2(e2, reader=None):
    """Pulls the vibrational data out of the file

    :param e2:
    :type e2:
    :return:
    :rtype:
    """

    # l = len(e2)
    # n = 1 + np.sqrt(1 + l/9) # I thought this was the way it was defined but...seems like not exactly
    n = reader.num_atoms
    # if n != int(n):
    #     n = l/14
    #     if n != int(n):
    #         raise ValueError("Gaussian FChk Vib-E2 block malformatted")
    # n = int(n)

    m = 3*n - 6

    freq = e2[:m]
    red_m = e2[m:2*m]
    frc_const = e2[2*m:3*m]
    intense = e2[3*m:4*m]

    return {
        "Frequencies"    : freq,
        "ReducedMasses"  : red_m,
        "ForceConstants" : frc_const,
        "Intensities"    : intense
    }
FormattedCheckpointComponents["Vib-E2"] = split_vib_e2

#endregion


########################################################################################################################
#
#                                           CommonNames
#

#region CommonNames

FormattedCheckpointCommonNames = {
    "Atomic numbers": "AtomicNumbers",
    "Current cartesian coordinates":"Coordinates",
    "Cartesian Gradient": "Gradient",
    "Cartesian Force Constants" : "ForceConstants",
    "Cartesian 3rd/4th derivatives" : "ForceDerivatives",
    "Dipole Moment" : "DipoleMoment",
    "Dipole Derivatives" : "DipoleDerivatives",
    "Dipole Moment num derivs" : "DipoleNumDerivatives",
    "Dipole Derivatives num derivs" : "DipoleHigherDerivatives",
    "Vib-E2" : "VibrationalData",
    "Vib-Modes" : "VibrationalModes",
    "Vib-AtMass" : "VibrationalAtomicMasses",
    "Real atomic weights" : "AtomicMasses"
}

#endregion


"""
Defines a SharedLibrary object that makes it relatively straightforward to use
extensions direclty from .so files
"""

import os, ctypes
from .ArgumentSignature import FunctionSignature, Argument

__all__ = [
    "SharedLibrary",
    "SharedLibraryFunction"
]

class SharedLibraryLoader:

    def __init__(self, shared_library):
        if not isinstance(shared_library, str):
            lib_file = shared_library._name # I'd prefer not to access private members but my options were limited
            if not os.path.isfile(lib_file):
                lib_file = None
        else:
            lib_file = shared_library
            shared_library = None
        self._lib = shared_library
        self._lib_file = lib_file

    class InDir:
        """
        A super simple context manager that manages going into a directory and then leaving when finished
        """

        def __init__(self, dir_name):
            self._to = dir_name
            self._from = None

        def __enter__(self):
            self._from = os.getcwd()
            os.chdir(self._to)

        def __exit__(self, exc_type, exc_val, exc_tb):
            if self._from is not None:
                os.chdir(self._from)
            self._from = None

    def in_dir(self):
        return self.InDir(self.lib_dir)

    @property
    def lib(self):
        if self._lib is None:
            with self.in_dir():
                self._lib = ctypes.cdll.LoadLibrary(self._lib_file)
        return self._lib

    @property
    def lib_dir(self):
        return os.path.dirname(self._lib_file)

class SharedLibraryFunction:
    """
    An object that provides a way to call into a shared library function
    """

    def __init__(self,
                 shared_library,
                 signature:FunctionSignature,
                 defaults=None,
                 docstring=None,
                 call_directory=None,
                 return_handler=None,
                 prep_args=None
                 ):
        """
        :param shared_library: the path to the shared library file you want to use
        :type shared_library: str |
        :param function_signature: the signature of the function to load
        :type function_signature: FunctionSignature
        :param call_directory: the directory for calling
        :type call_directory: str
        :param docstring: the docstring for the function
        :type docstring: str
        """
        if not isinstance(shared_library, SharedLibraryLoader):
            shared_library = SharedLibraryLoader(shared_library)
        self._loader = shared_library
        self._fun = None
        self._fname = None
        self._sig = signature
        self._doc = docstring
        # if call_directory is None:
        #     call_directory = self._loader.lib_dir
        self._dir = call_directory  # we could be kinder here and do stuff like add support for ".." and friends
        self.defaults = defaults
        if return_handler is None:
            return_handler = self._manage_return
        self.return_handler = return_handler
        self.arg_prepper = prep_args

    @classmethod
    def construct(cls,
                  name,
                  lib,
                  docstring=None,
                  defaults=None,
                  return_type=None,
                  return_handler=None,
                  **args
                  ):

        return cls(
            lib,
            FunctionSignature.construct(
                name
                **args,
                return_type=return_type
            ),
            docstring=docstring,
            defaults=defaults,
            return_handler=return_handler
        )

    @property
    def function(self):
        self.initialize()
        return self._fun
    def initialize(self):
        if self._fun is None:
            # means we need to load it from the shared lib
            if self._fname is None:
                self._fname = self._sig.name
            self._fun = getattr(self._loader.lib, self._fname)

            # now initialize the arg signature
            self._fun.restype = self._sig.return_type
            self._fun.argtypes = self._sig.arg_types # need to figure out what type I need...
    def doc(self):
        return self._sig.cpp_signature+"\n"+self._doc
    def __repr__(self):
        return "{}({}, {})".format(
            type(self).__name__,
            self._sig,
            self._loader.lib
        )

    @property
    def signature(self):
        return self._sig

    @classmethod
    def _manage_return(cls, res, args):
        return cls.uncast(res)
    @classmethod
    def uncast(cls, res):
        if hasattr(res, '_obj'): #byref
            res = res._obj
        if hasattr(res, 'value'):
            res = res.value
        return res

    def _call(self, args, kwargs): # here to be overloaded
        args = self._sig.prep_args(args, kwargs, defaults=self.defaults)
        if self._dir is not None:
            with SharedLibraryLoader.InDir(self._dir):
                res = self.function(*args)
        else:
            res = self.function(*args)
        args = dict(zip((a.name for a in self._sig.args), args))
        return res, args

    def call(self, *args, **kwargs):
        """
        Calls the function we loaded.
        This will be parallelized out to handle more complicated usages.

        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """

        if self.arg_prepper is not None:
            kwargs = self._sig.populate_kwargs(args, kwargs,  defaults=self.defaults)
            args = None
            kwargs = self.arg_prepper(kwargs)
        res, args = self._call(args, kwargs)
        return self.return_handler(res, args)

    def __call__(self, *args, **kwargs):
        return self.call(*args, **kwargs)

class SharedLibrary:

    method_type = SharedLibraryFunction
    def __init__(
            self,
            library,
            **functions
    ):
        if not isinstance(library, SharedLibraryLoader):
            library = SharedLibraryLoader(library)
        self._loader = library
        self._functions = {}
        for k,v in functions.items():
            self.register(k, **v)
    def register(self, tag, name=None, docstring=None, defaults=None, return_handler=None, prep_args=None, **params):
        if name is None:
            name = tag
        fn = self.method_type(
            self._loader,
            FunctionSignature.construct(name, **params),
            docstring=docstring,
            defaults=defaults,
            return_handler=return_handler,
            prep_args=prep_args
        )
        self._functions[tag] = fn
        return fn

    def get_function(self, item):
        if item in self._functions:
            return self._functions[item]
        else:
            raise ValueError("no shared library function {}".format(item))

    def __getattr__(self, item):
        return self.get_function(item)

    def __repr__(self):
        return "{}({})".format(
            type(self).__name__,
            ", ".format(repr(fn.sig) for fn in self._functions)
        )



__converters__ = []
from .CartesianToZMatrix import __converters__ as exposed
__converters__ += exposed
from .ZMatrixToCartesian import __converters__ as exposed
__converters__ += exposed
from .CartesianToSpherical import __converters__ as exposed
__converters__ += exposed
from .SphericalToCartesian import __converters__ as exposed
__converters__ += exposed
from .GenericInternalCoordinateSystem import __converters__ as exposed
__converters__ += exposed


import abc
import uuid

import numpy as np

from ..Jupyter import JHTML, X3DHTML
from ... import Numputils as nput

__all__ = [
    "X3D",
    "X3DPrimitive",
    "X3DGeometryObject",
    "X3DGeometryGroup",
    "X3DGroup",
    "X3DScene",
    "X3DMaterial",
    "X3DLine",
    "X3DSphere",
    "X3DCone",
    "X3DCylinder",
    "X3DTorus",
    "X3DSwitch",
    "X3DListAnimator"
]

class X3DObject(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def to_x3d(self):
        ...

class X3D(X3DObject):
    defaults = dict(
        width=500,
        height=500
    )
    def __init__(self, *children, **opts):
        if len(children) == 1 and isinstance(children[0], (tuple, list)):
            children = children[0]
        self.children = children
        self.opts = opts

    X3DOM_JS = 'http://www.x3dom.org/download/x3dom.js'
    X3DOM_CSS = 'http://www.x3dom.org/download/x3dom.css'
    def to_widget(self):
        return JHTML.Figure(
                JHTML.Script(src=self.X3DOM_JS),
                JHTML.Link(rel='stylesheet', href=self.X3DOM_CSS),
                #     "<script type='text/javascript' src='http://www.x3dom.org/download/x3dom.js'> </script>
                # <link rel='stylesheet' type='text/css' href='http://www.x3dom.org/download/x3dom.css'></link>"
                self.to_x3d()
            )
    def _ipython_repr_(self):
        return self.to_widget()
    def to_x3d(self):
        base_opts = dict(self.defaults, **self.opts)
        for k in ['width', 'height']:
            if k in base_opts:
                v = base_opts[k]
                if nput.is_numeric(v):
                    base_opts[k] = f'{v}px'
        return X3DHTML.X3D(
                    *[a.to_x3d() if hasattr(a, 'to_x3d') else a for a in self.children],
                    **base_opts
                )

class X3DMaterial(X3DObject):
    __props__ = {
        "diffuseColor",
        "ambientIntensity",
        "emissiveColor",
        "specularColor",
        "shininess",
        "transparency",
        # "metadata"
    }
    def __init__(self, **attrs):
        self.attrs = attrs

    conversion_map = {
        "brightness": "ambientIntensity",
        "glow":"emissiveColor",
        "color": "diffuseColor",
        "specularity": "specularColor"
    }
    @classmethod
    def prop_keys(cls):
        return (cls.__props__ | cls.conversion_map.keys())
    def prep_attrs(self, attrs:dict):
        attrs = {
            self.conversion_map.get(k, k):v
            for k,v in attrs.items()
        }
        excess_keys = attrs.keys() - self.__props__
        if len(excess_keys) > 0:
            raise ValueError(f"keys {excess_keys} are invalid material keys")
        return attrs
    def to_x3d(self):
        return X3DHTML.Material(**self.prep_attrs(self.attrs))

class X3DPrimitive(X3DObject):
    wrapper_class = None
    tag_class = None
    def __init__(self, *children, **opts):
        if len(children) == 1 and isinstance(children[0], (tuple, list)):
            children = children[0]
        self.children = children
        self.opts = opts
    def split_opts(self, opts:dict):
        material_keys = opts.keys() & X3DMaterial.prop_keys()
        rem_keys = opts.keys() - material_keys
        return {k:opts[k] for k in rem_keys}, {k:opts[k] for k in material_keys}
    def get_appearance(self, material_opts):
        if len(material_opts) == 0:
            return None
        else:
            return X3DHTML.Appearance(X3DMaterial(**material_opts).to_x3d())
    def to_x3d(self):
        obj_opts, material_opts = self.split_opts(self.opts)
        kids = [k.to_x3d() if hasattr(k, 'to_x3d') else k for k in self.children]
        appearance = self.get_appearance(material_opts)
        if self.tag_class is None:
            if appearance is not None:
                kids = [appearance] + kids
            return self.wrapper_class(
                kids,
                **obj_opts
            )
        else:
            core = self.tag_class(kids, **obj_opts)
            appearance = self.get_appearance(material_opts)
            if appearance is not None:
                return self.wrapper_class(appearance, core)
            else:
                return core

class X3DScene(X3DPrimitive):
    wrapper_class = X3DHTML.Scene
    default_viewpoint = {'viewAll':True}
    def __init__(self, *children, viewpoint=None, **opts):
        if viewpoint is None:
            viewpoint = self.default_viewpoint
        elif viewpoint is False:
            viewpoint = {}
        super().__init__(*children, **opts)
        if len(viewpoint) > 0:
            self.children = [X3DHTML.Viewpoint(**viewpoint)] + list(self.children)

class X3DGroup(X3DPrimitive):
    wrapper_class = X3DHTML.Group

class X3DSwitch(X3DPrimitive):
    wrapper_class = X3DHTML.Switch

class X3DGeometryObject(X3DPrimitive):
    wrapper_class = X3DHTML.Shape
    def __init__(self, *args, **opts):
        geom_opts, self.material_opts = self.split_opts(opts)
        self.geometry_opts = self.prep_geometry_opts(*args, **geom_opts)
        super().__init__()
    @abc.abstractmethod
    def prep_geometry_opts(self, *args, **opts) -> dict:
        ...
    def create_object(self, translation=None, rotation=None, scale=None, **core_opts):
        base_obj = self.tag_class(**core_opts)
        tf = {}
        for k,v in [["translation",translation], ["rotation",rotation], ["scale", scale]]:
            if v is not None:
                tf[k] = v
        if len(tf) == 0:
            tf = None
        # base_obj = X3DHTML.Transform(base_obj, translation=translation, rotation=rotation, scale=scale)
        return base_obj, tf
    def get_rotation(self, axis):
        angs, crosses, norms = nput.vec_angles([0, 1, 0], axis, return_crosses=True, return_norms=True)
        if nput.is_numeric(angs):
            return np.concatenate([crosses, [angs]]), norms[1]
        else:
            return np.concatenate([crosses, angs[..., np.newaxis]], axis=-1), norms[1]
    def to_x3d(self):
        # obj_opts, material_opts = self.split_opts(self.opts)
        # kids = [k.to_x3d() for k in self.children]
        core, tf = self.create_object(**self.geometry_opts)
        appearance = self.get_appearance(self.material_opts)
        if appearance is not None:
            core = self.wrapper_class(appearance, core)
        if tf is not None:
            core = X3DHTML.Transform(core, **tf)
        return core

class X3DGeometryGroup(X3DGeometryObject):
    @abc.abstractmethod
    def prep_geometry_opts(self, *args, **opts) -> list[dict]:
        ...
    def prep_vecs(self, vecs, nstruct=None):
        if vecs is None:
            return [None] * nstruct
        else:
            vecs = np.asanyarray(vecs)
            if vecs.ndim == 1:
                vecs = vecs[np.newaxis]
            if nstruct is not None:
                vecs = np.broadcast_to(vecs, (nstruct, vecs.shape[-1]))
        return vecs
    def prep_mats(self, mats, nstruct=None):
        if mats is None:
            return [None] * nstruct
        else:
            mats = np.asanyarray(mats)
            if mats.ndim == 2:
                mats = mats[np.newaxis]
            if nstruct is not None:
                mats = np.broadcast_to(mats, (nstruct,) + mats.shape[-2:])
            return mats
    def prep_const(self, const, nstruct):
        if const is None:
            return [None] * nstruct
        else:
            const = np.asanyarray(const)
            if const.ndim == 0:
                const = const[np.newaxis]
            const = np.broadcast_to(const, (nstruct,))
            return const
    def to_x3d(self):
        kids = [self.create_object(**g) for g in self.geometry_opts]
        appearance = self.get_appearance(self.material_opts)
        objs = []
        for o,tf in kids:
            if appearance is not None:
                o = self.wrapper_class(appearance, o)
            if tf is not None:
                o = X3DHTML.Transform(o, **tf)
            objs.append(o)
        if len(objs) == 1:
            return objs[0]
        else:
            return X3DHTML.Group(objs)

class X3DLine(X3DGeometryObject):
    tag_class = X3DHTML.LineSet

    def prep_geometry_opts(self, points, **opts)->dict:
        return {"coords":np.asanyarray(points).flatten()}

class X3DSphere(X3DGeometryGroup):
    tag_class = X3DHTML.Sphere

    def prep_geometry_opts(self, centers, radius=1, **opts):
        centers = self.prep_vecs(centers)
        rads = self.prep_const(radius, centers.shape[0])
        return [{"translation":c, "radius":r} for c,r in zip(centers, rads)]

class X3DCylinder(X3DGeometryGroup):
    tag_class = X3DHTML.Cylinder

    def prep_geometry_opts(self, starts, ends, radius=1, **opts):
        starts = self.prep_vecs(starts)
        ends = self.prep_vecs(ends)
        radius = self.prep_const(radius, starts.shape[0])

        axes = ends - starts
        rots, norms = self.get_rotation(axes)

        return [
            {"translation":s, "rotation":a, "height":n, "radius":r}
            for s,a,n,r in zip((starts + ends) / 2, rots, norms, radius)
        ]

class X3DCone(X3DGeometryGroup):
    tag_class = X3DHTML.Cone

    def prep_geometry_opts(self, starts, ends, radius=1, top_radius=None, **opts):
        starts = self.prep_vecs(starts)
        ends = self.prep_vecs(ends)
        radius = self.prep_const(radius, starts.shape[0])
        top_radius = self.prep_const(top_radius, starts.shape[0])

        axes = ends - starts
        rots, norms = self.get_rotation(axes)

        return [
            {"translation":s, "rotation":a, "height":n, "bottomRadius":r, "topRadius":t}
            for s,a,n,r,t in zip((starts + ends) / 2, rots, norms, radius, top_radius)
        ]

class X3DTorus(X3DGeometryGroup):
    tag_class = X3DHTML.Torus

    def prep_geometry_opts(self, centers, radius=1, inner_radius=None, **opts):
        centers = self.prep_vecs(centers)
        radius = self.prep_const(radius, centers.shape[0])
        inner_radius = self.prep_const(inner_radius, centers.shape[0])

        return [
            {"translation":s,  "outerRadius":r, "innerRadius":i}
            for s,r,i in zip(centers, radius, inner_radius)
        ]

class X3DListAnimator(X3DGroup):
    def __init__(self, *frames, id=None, animation_duration=2, running=True, slider=False, **opts):
        self.uuid = str(uuid.uuid4())
        if id is None:
            id = f"animation-{self.uuid}"
        self.id = id
        anim_frames = X3DSwitch(
                *frames,
                id=id,
                whichChoice="0"
            )
        nframes = len(anim_frames.children) - 1
        key_frames = np.linspace(0, 1, nframes)
        elements = []
        if slider:
            elements.append(
                JHTML.Input(type="range", value="0", min="0", max=f"{nframes}", step="1", cls="slider",
                            oninput=f"""document.getElementById("{id}").setAttribute("whichChoice", this.value)""")
            )
        elements.append(anim_frames)
        if running:
            elements.extend([
                X3DHTML.TimeSensor(id=f'animation-clock-{self.uuid}', cycleInterval=animation_duration, loop=True,
                                   enabled=running
                                   ),
                X3DHTML.IntegerSequencer(id=f'animation-indexer-{self.uuid}', key=key_frames,
                                         keyValue=np.arange(nframes)),
                X3DHTML.Route(
                    fromField='fraction_changed', fromNode=f'animation-clock-{self.uuid}',
                    toField='set_fraction', toNode=f'animation-indexer-{self.uuid}'
                ),
                X3DHTML.Route(
                    fromField='value_changed', fromNode=f'animation-indexer-{self.uuid}',
                    toField='whichChoice', toNode=self.id
                )
            ])

        super().__init__(elements, **opts)




"""
This lists the types of readers and things available to the GaussianLogReader
"""
import numpy as np

from ..Parsers import *
from collections import namedtuple, OrderedDict

########################################################################################################################
#
#                                           GaussianLogComponents
#
# region GaussianLogComponents
GaussianLogComponents = OrderedDict()  # we'll register on this bit by bit
# each registration should look like:

# GaussianLogComponents["Name"] = {
#     "description" : string, # used for docmenting what we have
#     "tag_start"   : start_tag, # starting delimeter for a block
#     "tag_end"     : end_tag, # ending delimiter for a block None means apply the parser upon tag_start
#     "parser"      : parser, # function that'll parse the returned list of blocks (for "List") or block (for "Single")
#     "mode"        : mode # "List" or "Single"
# }

########################################################################################################################
#
#                                           Header
#

tag_start = "******************************************"
tag_end   = FileStreamerTag(
    """ --------""",
    follow_ups = (""" -----""",)
)

HeaderPercentBlockParser = StringParser(
    NonCapturing(
        ("%", Capturing(Word, dtype=str), "=", Capturing((Word, Optional(Any), Word), dtype=str) ),
        dtype=str
    )
)
HeaderHashBlockLine = RegexPattern((
    Capturing(Optional((Word, "="), dtype=str)),
    Capturing(Repeating(Alternatives((WordCharacter, "\(", "\)", "\/", "\-")), dtype=str))
    ))
HeaderHashBlockLineParser = StringParser(HeaderHashBlockLine)
HeaderHashBlockParser = StringParser(
    RegexPattern((
        "#",
         Optional(Whitespace),
         Repeating(
             Named(
                Repeating(HeaderHashBlockLine),
                "Command",
                suffix=Optional(Whitespace),
                dtype=str,
                default=""
             )
        )
    ))
)

def header_parser(header):
    # regex = HeaderHashBlockParser.regex #type: RegexPattern
    header_percent_data = HeaderPercentBlockParser.parse_all(header)
    runtime_options = {}
    for k, v in header_percent_data.array:
        runtime_options[k] = v

    header_hash_data = HeaderHashBlockParser.parse_all(header)
    all_keys=" ".join(header_hash_data["Command"].array.flatten())
    raw_data = HeaderHashBlockLineParser.parse_all(all_keys).array

    job_options = {}
    for k, v in raw_data:
        if k.endswith("="):
            job_options[k.strip("=").lower()] = v.strip("()").split(",")
        else:
            job_options[v] = []

    return namedtuple("HeaderData", ["config", 'job'])(runtime_options, job_options)

mode = "Single"

GaussianLogComponents["Header"] = {
    "tag_start": tag_start,
    "tag_end"  : tag_end,
    "parser"   : header_parser,
    "mode"     : mode
}

########################################################################################################################
#
#                                           InputZMatrix
#

# region InputZMatrix
tag_start = "Z-matrix:"
tag_end   = """ 
"""

def parser(zmat):
    return zmat

mode = "Single"

GaussianLogComponents["InputZMatrix"] = {
    "tag_start": tag_start,
    "tag_end"  : tag_end,
    "parser"   : parser,
    "mode"     : mode
}

# endregion

########################################################################################################################
#
#                                           CartesianCoordinates
#

# region CartesianCoordinates

 # the region thing is just a PyCharm hack to collapse the boilerplate here... Could also have done 5000 files

cart_delim = """ --------------------------------------------------------------"""
cartesian_start_tag = FileStreamerTag(
    """Center     Atomic      Atomic             Coordinates (Angstroms)""",
    follow_ups= cart_delim
)
cartesian_end_tag = cart_delim

CartParser = StringParser(
    Repeating(
        (
            Named(
                Repeating(
                    Capturing(PositiveInteger),
                    min=3, max=3,
                    prefix=Optional(Whitespace),
                    suffix=Whitespace
                ),
                "GaussianStuff", handler=StringParser.array_handler(dtype=int)
            ),
            Named(
                Repeating(
                    Capturing(Number),
                    min = 3,
                    max = 3,
                    prefix=Optional(Whitespace),
                    joiner = Whitespace
                ),
                "Coordinates", handler=StringParser.array_handler(dtype=float)
            )
        ),
        suffix = Optional(Newline)
    )
)

# raise Exception(CartParser.regex)

def cartesian_coordinates_parser(strs):
    strss = "\n\n".join(strs)

    parse = CartParser.parse_all(strss)

    coords = (
        parse["GaussianStuff", 0],
        parse["Coordinates"].array
    )

    return coords

GaussianLogComponents["CartesianCoordinates"] = {
    "tag_start": cartesian_start_tag,
    "tag_end"  : cartesian_end_tag,
    "parser"   : cartesian_coordinates_parser,
    "mode"     : "List"
}
GaussianLogComponents["ZMatCartesianCoordinates"] = {
    "tag_start": FileStreamerTag('''Z-Matrix orientation:''', follow_ups = (cart_delim, cart_delim)),
    "tag_end"  : cartesian_end_tag,
    "parser"   : cartesian_coordinates_parser,
    "mode"     : "List"
}
GaussianLogComponents["StandardCartesianCoordinates"] = {
    "tag_start": FileStreamerTag('''Standard orientation:''', follow_ups = (cart_delim, cart_delim)),
    "tag_end"  : cartesian_end_tag,
    "parser"   : cartesian_coordinates_parser,
    "mode"     : "List"
}
GaussianLogComponents["InputCartesianCoordinates"] = {
    "tag_start": FileStreamerTag('''Input orientation:''', follow_ups = (cart_delim, cart_delim)),
    "tag_end"  : cartesian_end_tag,
    "parser"   : cartesian_coordinates_parser,
    "mode"     : "List"
}

# endregion

########################################################################################################################
#
#                                           ZMatrices
#

# region ZMatrices
tag_start = """Z-MATRIX (ANGSTROMS AND DEGREES)
   CD    Cent   Atom    N1       Length/X        N2       Alpha/Y        N3        Beta/Z          J
 ---------------------------------------------------------------------------------------------------"""
tag_end   = " ---------------------------------------------------------------------"

ZMatParser = StringParser(
    Repeating(
        (
            Named(
                Repeating(Capturing(PositiveInteger), min=1, max=2, prefix=Optional(Whitespace), suffix=Whitespace),
                "GaussianInts"
            ),
            Named(
                Capturing(AtomName),
                "AtomNames",
                suffix=Whitespace
            ),
            Named(
                Repeating(
                    (
                        Capturing(PositiveInteger),
                        Capturing(Number),
                        Parenthesized(PositiveInteger, prefix=Whitespace)
                    ),
                    min = None,
                    max = 3,
                    prefix=Optional(Whitespace),
                    joiner = Whitespace,
                ),
                "Coordinates"
            )
        ),
        suffix = Optional(RegexPattern((
            Optional((Whitespace, PositiveInteger)),
            Optional(Newline)
        )))
    )
)

def parser(strs):

    strss = '\n\n'.join(strs)
    fak = ZMatParser.parse_all(strss)
    coords = [
        (
            fak["GaussianInts", 0],
            fak["AtomNames", 0]
        ),
        fak["Coordinates", 0, 0],
        fak["Coordinates", 1]
        ]

    return coords
mode = "List"

GaussianLogComponents["ZMatrices"] = {
    "tag_start": tag_start,
    "tag_end"  : tag_end,
    "parser"   : parser,
    "mode"     : mode
}

# endregion

########################################################################################################################
#
#                                           OptimizationParameters
#

# region OptimizationParameters

tag_start  = "Optimization "
tag_end    = """                        !
 ------------------------------------------------------------------------
"""


def parser(pars):
    """Parses a optimizatioon parameters block"""
    did_opts = [ "Non-Optimized" not in par for par in pars]
    return did_opts, pars


mode = "List"

GaussianLogComponents["OptimizationParameters"] = {
    "tag_start": tag_start,
    "tag_end"  : tag_end,
    "parser"   : parser,
    "mode"     : mode
}

# endregion

########################################################################################################################
#
#                                           MullikenCharges
#

#region MullikenCharges
tag_start = "Mulliken charges:"
tag_end   = "Sum of Mulliken charges"


def parser(charges):
    """Parses a Mulliken charges block"""
    return charges
mode = "List"

GaussianLogComponents["MullikenCharges"] = {
    "tag_start": tag_start,
    "tag_end"  : tag_end,
    "parser"   : parser,
    "mode"     : mode
}

# endregion

########################################################################################################################
#
#                                           MultipoleMoments
#

#region MultipoleMoments
tag_start  = "Dipole moment ("
tag_end    = " N-N="


def parser(moms):
    """Parses a multipole moments block"""
    return moms


mode = "List"

GaussianLogComponents["MultipoleMoments"] = {
    "tag_start": tag_start,
    "tag_end"  : tag_end,
    "parser"   : parser,
    "mode"     : mode
}

# endregion

########################################################################################################################
#
#                                           DipoleMoments
#

# region DipoleMoments
tag_start  = "Dipole moment ("
tag_end    = "Quadrupole moment ("

dips_parser = StringParser(
    RegexPattern(
        (
            "X=", Capturing(Number),
            "Y=", Capturing(Number),
            "Z=", Capturing(Number)
        ),
        joiner=Whitespace,
        dtype=(float, (3,))
    )
)
def parser(moms):
    """Parses a multipole moments block"""

    res = dips_parser.parse_all("\n".join(moms))
    return res.array

mode = "List"
GaussianLogComponents["DipoleMoments"] = {
    "tag_start" : tag_start,
    "tag_end"   : tag_end,
    "parser"    : parser,
    "mode"      : mode
}

# endregion

########################################################################################################################
#
#                                           OptimizedDipoleMoments
#

# region DipoleMoments
tag_start  = " Dipole        ="
tag_end    = " Optimization"


def convert_D_number(a, **kw):
    import numpy as np
    return np.array([float(s.replace("D", "E")) for s in a])
DNumberPattern = RegexPattern((Number, "D", Integer), dtype=float)
OptimizedDipolesParser = StringParser(
    RegexPattern(
        (
            "Dipole", "=",
            Repeating(
                Capturing(DNumberPattern, handler=convert_D_number),
                min=3,
                max=3,
                suffix=Optional(Whitespace)
            )
        ),
        joiner=Whitespace
    )
)

def parser(mom):
    """Parses dipole block, but only saves the dipole of the optimized structure"""
    import numpy as np

    mom = "Dipole  =" + mom
    # print(">>>>>", mom)
    grps = OptimizedDipolesParser.parse_iter(mom)
    match = None
    for match in grps:
        pass

    if match is None:
        return np.array([])
    return match.value.array
    # else:
    #     grp = match.value
    #     dip_list = [x.replace("D", "E") for x in grp]
    #     dip_array = np.asarray(dip_list)
    #     return dip_array.astype("float64")

mode       = "List"
parse_mode = "Single"

GaussianLogComponents["OptimizedDipoleMoments"] = {
    "tag_start" : tag_start,
    "tag_end"   : tag_end,
    "parser"    : parser,
    "mode"      : mode,
    "parse_mode": parse_mode
}

# endregion

########################################################################################################################
#
#                                           ScanEnergies
#

# region ScanEnergies

tag_start = """ Summary of the potential surface scan:"""
tag_end = """Normal termination of"""

# Number = '(?:[\\+\\-])?\\d*\\.\\d+'
# block_pattern = "\s*"+Number+"\s*"+Number+"\s*"+Number+"\s*"+Number+"\s*"+Number
# block_re = re.compile(block_pattern)

ScanEnergiesParser = StringParser(
    RegexPattern(
        (
            Named(
                Repeating(Capturing(Word), prefix=Whitespace),
                "Keys",
                suffix=NonCapturing([Whitespace, Newline])
            ),
            Named(
                Repeating(
                    Repeating(
                        Alternatives([PositiveInteger, Number], dtype=float),
                        prefix=Whitespace
                        ),
                    suffix=Newline
                ),
                "Coords",
                prefix=Newline,
                handler=StringParser.array_handler(dtype=float)
            ),
        ),
        joiner=NonCapturing(Repeating([Whitespace, Repeating(["-"])]))
    )
)

def parser(block):
    """Parses the scan summary block"""
    import re

    if block is None:
        raise KeyError("key '{}' not in .log file".format('ScanEnergies'))

    r = ScanEnergiesParser.regex # type: RegexPattern
    parse=ScanEnergiesParser.parse(block)
    
    return namedtuple("ScanEnergies", ["coords", "energies"])(
        parse["Keys"].array,
        parse["Coords"].array
    )

mode = "Single"

GaussianLogComponents["ScanEnergies"] = {
    "tag_start": tag_start,
    "tag_end"  : tag_end,
    "parser"   : parser,
    "mode"     : mode
}

# endregion

########################################################################################################################
#
#                                           OptimizedScanEnergies
#

# region OptimizedScanEnergies

tag_start = """ Summary of Optimized Potential Surface Scan"""
tag_end = FileStreamerTag(
    tag_alternatives = (
        """ Largest change from initial coordinates is atom """,
        """-"""*25
    )
)

eigsPattern = RegexPattern(
    (
        "Eigenvalues --",
        Repeating(Capturing(Number), suffix=Optional(Whitespace))
    ),
    joiner=Whitespace
)

coordsPattern = RegexPattern(
    (
        Capturing(VariableName),
        Repeating(Capturing(Number), suffix=Optional(Whitespace))
    ),
    prefix=Whitespace,
    joiner=Whitespace
)

OptScanPat = StringParser(
    RegexPattern(
        (
            Named(eigsPattern,
                  "Eigenvalues"
                  #parser=lambda t: np.array(Number.findall(t), 'float')
                  ),
            Named(Repeating(coordsPattern, suffix=Optional(Newline)), "Coordinates")
        ),
        joiner=Newline
    )
)

# Gaussian16 started subtracting any uniform shift off of the energies
# we'd like to get it back

eigsShift = RegexPattern(
    (
        "add",
        Whitespace,
        Named(Number, "Shift")
    ),
    joiner=Whitespace
)

EigsShiftPat = StringParser(eigsShift)

def parser(pars):
    """Parses the scan summary block and returns only the energies"""
    from collections import OrderedDict
    import numpy as np

    if pars is None:
        return None

    try:
        shift = EigsShiftPat.parse(pars)["Shift"]
        #TODO: might need to make this _not_ be a `StructuredTypeArray` at some point, but seems fine for now
    except StringParserException:
        shift = 0
    par_data = OptScanPat.parse_all(pars)
    energies_array = np.concatenate(par_data["Eigenvalues"]).flatten()\
    # when there isn't a value, for shape reasons we get extra nans
    energies_array = energies_array[np.logical_not(np.isnan(energies_array))] + shift
    coords = OrderedDict()
    cdz = [a.array for a in par_data["Coordinates"].array]
    for coord_names, coord_values in zip(*cdz):
        for k, v in zip(coord_names, coord_values):
            if k not in coords:
                coords[k] = [v]
            else:
                coords[k].append(v)
    for k in coords:
        coords[k] = np.concatenate(coords[k]).flatten()
        coords[k] = coords[k][np.logical_not(np.isnan(coords[k]))]

    return namedtuple("OptimizedScanEnergies", ["energies", "coords"])(
        energies_array,
        coords
    )

mode = "Single"

GaussianLogComponents["OptimizedScanEnergies"] = {
    "tag_start": tag_start,
    "tag_end"  : tag_end,
    "parser"   : parser,
    "mode"     : mode
}

# endregion

########################################################################################################################
#
#                                           X-Matrix
#

# region X-Matrix

tag_start = FileStreamerTag(
    """Total Anharmonic X Matrix (in cm^-1)""",
    follow_ups=("""-"""*25,)
)
tag_end = FileStreamerTag(
    tag_alternatives = (
        """ ================================================== """,
        """-"""*25
    )
)

def parser(pars):
    """Parses the X matrix block and returns stuff --> huge pain in the ass function"""
    import numpy as np

    energies = np.array([x.replace("D", "E") for x in DNumberPattern.findall(pars)])
    l = len(energies)
    n = int( (-1 + np.sqrt(1 + 8*l))/2 )
    X = np.empty((n, n))
    # gaussian returns the data as blocks of 5 columns in the lower-triangle, annoyingly,
    # so we need to rearrange the indices so that they are sorted to make this work
    i, j = np.tril_indices_from(X)
    energies_taken = 0
    blocks = int(np.ceil(n/5))
    for b in range(blocks):
        sel = np.where((b*5-1 < j) * (j < (b+1)*5))[0]
        e_new=energies_taken+len(sel)
        e = energies[energies_taken:e_new]
        energies_taken=e_new
        ii = i[sel]
        jj = j[sel]
        X[ii, jj] = e
        X[jj, ii] = e

    return X

mode = "Single"

GaussianLogComponents["XMatrix"] = {
    "tag_start": tag_start,
    "tag_end"  : tag_end,
    "parser"   : parser,
    "mode"     : mode
}

# endregion


tag_start =  "Job cpu time"
tag_end = "Normal termination"

def parser(block, start=tag_start):
    return " " + start + block

mode = "Single"

GaussianLogComponents["Footer"] = {
    "tag_start": tag_start,
    "tag_end"  : tag_end,
    "parser"   : parser,
    "mode"     : mode
}

# endregion

tag_start = """FrcOut:"""
tag_end = """MW cartesian velocity:"""

CartesianBlockTags = ["Cartesian coordinates: (bohr)", "MW cartesian"]
def parse_aimd_coords(blocks):
    big_block = "\n".join(blocks)
    comps = np.char.replace(DNumberPattern.findall(big_block), "D", "E").astype(float)
    return comps.reshape((len(blocks), -1, 3)) # easy as that since XYZ?

ValueBlockTags = ["FrcOut",
                  FileStreamerTag(
                      """Final forces over variables""",
                      follow_ups=("Leave Link",)
                  )]
def convert_D_number(a, **kw):
    import numpy as np
    converted = np.char.replace(a, 'D', 'E')
    return converted.astype(float)
DNumberPattern = RegexPattern((Number, "D", Integer), dtype=float)
EnergyBlockPattern = StringParser(
        RegexPattern(
            (
                "Energy=", Named(DNumberPattern, 'E', handler=convert_D_number)
            )
        )
)
ForceBlockTags = ["force vector number 2", "After rot"]
def parse_grad(block):
    comps = np.array([x.replace("D", "E") for x in DNumberPattern.findall(block)])
    return comps.astype(float) #.reshape((-1, 3)) # easy as that since XYZ? -> even easier...
def parse_weird_mat(pars): # identical to X-matrix parser...
    """Parses the Hessian matrix block and returns stuff --> huge pain in the ass function"""
    import numpy as np

    energies = np.array([x.replace("D", "E") for x in DNumberPattern.findall(pars)])
    l = len(energies)
    n = int( (-1 + np.sqrt(1 + 8*l))/2 )
    X = np.empty((n, n))
    # gaussian returns the data as blocks of 5 columns in the lower-triangle, annoyingly,
    # so we need to rearrange the indices so that they are sorted to make this work
    i, j = np.tril_indices_from(X)
    energies_taken = 0
    blocks = int(np.ceil(n/5))
    for b in range(blocks):
        sel = np.where((b*5-1 < j) * (j < (b+1)*5))[0]
        e_new=energies_taken+len(sel)
        e = energies[energies_taken:e_new]
        energies_taken=e_new
        ii = i[sel]
        jj = j[sel]
        X[ii, jj] = e
        X[jj, ii] = e

    return X
HessianBlockTags = ["Force constants in Cartesian coordinates:", "Final forces"]


tag_start = """Full mass-weighted force constant matrix:"""
tag_end = """
 
"""
"""
                    193                    194                    195
                      A                      A                      A
 Frequencies --   3235.3026              3250.8741              3253.7284
 Red. masses --      1.0940                 1.0918                 1.0944
 Frc consts  --      6.7468                 6.7982                 6.8264
 IR Inten    --      5.0789                 1.9672                 1.1082
  Atom  AN      X      Y      Z        X      Y      Z        X      Y      Z
     1   8     0.00  -0.00   0.00     0.00  -0.00   0.00     0.00   0.00  -0.00
 """

normal_mode_block = RegexPattern(
    (
        Named(Repeating((Whitespace, Integer), min=1, max=3), "label"),
        Named(Repeating((Whitespace, Word), min=1, max=3), "symmetry"),
        Named(("Frequencies --", Repeating((Whitespace, Number), min=1, max=3)), "freqs"),
        Named(("Red. masses --", Repeating((Whitespace, Number), min=1, max=3)), "masses"),
        Named(("Frc consts  --", Repeating((Whitespace, Number), min=1, max=3)), "fcs"),
        Named(("IR Inten    --", Repeating((Whitespace, Number), min=1, max=3)), "ir_ints"),
        RegexPattern(("Atom  AN", Repeating((Whitespace, "X      Y      Z"), min=1, max=3))),
        Named(
            Repeating(
                (Whitespace, Integer, Whitespace, Integer, Repeating((Whitespace, Number), min=3, max=9), Newline)
            ),
            "displacements"
        )
    ),
    joiner=RegexPattern((Newline, Whitespace))
)

def parse_nms_modes(label, symmetries, freqs, masses, fcs, ir_ints, _, disps):
    freqs = [float(x) for x in freqs[len('Frequencies --'):].split()] # since len max 3, probably fastest way
    masses = [float(x) for x in masses[len('Red. masses --'):].split()] # since len max 3, probably fastest way
    disps = StringParser.array_handler(dtype=float)(disps)[:, 2:]
    disps = np.moveaxis(disps.reshape(disps.shape[0], 3, -1), 1, -1).reshape(-1, 3)
    return freqs, masses, disps
def parse_nms_block(block):
    bits = normal_mode_block.findall(block)
    freqs = [None] * len(bits)
    masses = [None] * len(bits)
    disps = [None] * len(bits)
    for i,bit in enumerate(bits):
        f, m, d = parse_nms_modes(*bit)
        freqs[i] = f
        masses[i] = m
        disps[i] = d

    return np.concatenate(freqs), np.concatenate(masses), np.concatenate(disps, axis=1)
def parse(blocks):
    return [parse_nms_block(b) for b in blocks]

mode = "List"
GaussianLogComponents["NormalModes"] = {
    "tag_start": tag_start,
    "tag_end"  : tag_end,
    "parser"   : parse,
    "mode"     : mode
}


def convert_D_number_block(a, **kw):
    import numpy as np
    convertable = np.char.replace(a.view('U15').reshape(a.shape[0], -1), 'D', 'E')
    return convertable.astype(float)
DNumberPattern = RegexPattern((Number, "D", Integer), dtype=float)
DipoleBlockPattern = StringParser(
        RegexPattern(
            (
                "Dipole", Whitespace, "=",
                    Named(Repeating(DNumberPattern, min=3, max=3), 'D', handler=convert_D_number_block),
            )
        )
)

def parse_aimd_values_blocks(blocks):
    big_block = "\n".join(blocks)

    energies = EnergyBlockPattern.parse_all(big_block)['E'].array
    # if energies[-1] == energies[-2]:
    #     if blocks[-1] == blocks[-2]:
    #         big_block = "\n".join(blocks[:-1]) # there's an extra copy
    #     energies = energies[:-1]

    # dips = DipoleBlockPattern.parse_all(big_block)['D'].array
    with StringStreamReader(big_block) as subparser:
        grad = np.array(subparser.parse_key_block(
            ForceBlockTags[0],
            ForceBlockTags[1],
            parser=lambda hstack: [parse_grad(h) for h in hstack],
            mode='List'
        ))

        hesses = np.array(subparser.parse_key_block(
            HessianBlockTags[0],
            HessianBlockTags[1],
            parser=lambda hstack:[parse_weird_mat(h) for h in hstack],
            mode='List'
        ))

    return namedtuple("AIMDValues", ['energies', 'gradients', 'hessians'])(
        energies=energies, gradients=grad, hessians=hesses#, dipoles=dips
    )

def parser(blocks):
    big_block = "FrcOut:\n".join([""] + [b.split('FrcOut:')[-1] for b in blocks]) # we can get two energy prints to one coords

    with StringStreamReader(big_block) as subparser:
        coords = subparser.parse_key_block(
            CartesianBlockTags[0],
            CartesianBlockTags[1],
            parser=parse_aimd_coords,
            mode='List'
        )

        vals = subparser.parse_key_block(
            ValueBlockTags[0],
            ValueBlockTags[1],
            parser=parse_aimd_values_blocks,
            mode='List'
        )

    return namedtuple("AIMDTrajectory", ['coords', 'vals'])(
        coords=coords,
        vals=vals
    )

mode = "List"
GaussianLogComponents["AIMDTrajectory"] = {
    "tag_start": tag_start,
    "tag_end"  : tag_end,
    "parser"   : parser,
    "mode"     : mode
}

# mode = "List"
# GaussianLogComponents["AIMDCoordinates"] = {
#     "tag_start": tag_start,
#     "tag_end"  : tag_end,
#     "parser"   : parser,
#     "mode"     : mode
# }

########################################################################################################################
#
#                                           GaussianLogDefaults
#
# region GaussianLogDefaults
GaussianLogDefaults = (
    "StartDateTime",
    "InputZMatrix",
    "ScanTable",
    "Blurb",
    "ComputerTimeElapsed",
    "EndDateTime"
)
# endregion

########################################################################################################################
#
#                                           GaussianLogOrdering
#
# region GaussianLogOrdering
# defines the ordering in a GaussianLog file
glk = ( # this must be sorted by what appears when
    "Header",
    "StartDateTime",
    "CartesianCoordinates",
    "ZMatCartesianCoordinates",
    "StandardCartesianCoordinates",
    "CartesianCoordinateVectors",
    "MullikenCharges",
    "MultipoleMoments",
    "DipoleMoments",
    "OptimizedDipoleMoments",
    "QuadrupoleMoments",
    "OctapoleMoments",
    "HexadecapoleMoments",
    "IntermediateEnergies",
    "InputZMatrix",
    "InputZMatrixVariables",
    "ZMatrices",
    "ScanEnergies",
    "OptimizedScanEnergies",
    "OptimizationScan",
    "Blurb",
    "Footer"
)
list_type = { k:-1 for k in GaussianLogComponents if GaussianLogComponents[k]["mode"] == "List" }
GaussianLogOrdering = { k:i for i, k in enumerate([k for k in glk if k not in list_type]) }
GaussianLogOrdering.update(list_type)
del glk
del list_type
# endregion




import abc, numpy as np, itertools, collections, copy
from functools import reduce

# from ...Combinatorics import SymmetricGroupGenerator

from  ...Misc import Abstract


__all__ = [
    "Symbols",
    "SymPyFunction"
    # "Summation",
    # "Product"
]

class Functionlike(metaclass=abc.ABCMeta):
    """
    A function suitable for symbolic manipulation
    with derivatives and evlauation
    """

    @abc.abstractmethod
    def eval(self, r: np.ndarray) -> 'np.ndarray':
        ...

    compile_vars = 0
    @staticmethod
    def cur_var():
        return Functionlike.compile_vars
    @staticmethod
    def inc_var():
        v = Functionlike.compile_vars
        Functionlike.compile_vars += 1
        return v
    @staticmethod
    def reset_var():
        Functionlike.compile_vars = 0
    @staticmethod
    def get_compile_var():
        return Abstract.Name("var_"+str(Functionlike.inc_var()))
    def get_compile_spec(self)->'Abstract.Expr':
        raise NotImplementedError("don't know how to compile {}".format(self))
    # def get_eval_ast(self)->'ast.Lambda':
    #     return ast.fix_missing_locations(ast.Lambda(
    #             args=ast.arguments(
    #                 posonlyargs=[],
    #                 args=[ast.arg(arg='x', annotation=None, type_comment=None)],
    #                 vararg=None,
    #                 kwonlyargs=[], kw_defaults=[], kwarg=None, defaults=[]
    #             ),
    #             body=self.get_compile_spec()
    #         ))
    # def lambda_wrap(self, var, expr)->'ast.Lambda':
    #     return ast.Lambda(
    #         args=ast.arguments(
    #             posonlyargs=[],
    #             args=[ast.arg(arg=var, annotation=None, type_comment=None)],
    #             vararg=None,
    #             kwonlyargs=[], kw_defaults=[], kwarg=None, defaults=[]
    #         ),
    #         body=expr
    #     )
    # @staticmethod
    # def ast_var(var):
    #     return ast.Name(id=var, ctx=ast.Load())

    # @staticmethod
    # def ast_call(fn, *args, **kwargs):
    #     return ast.Call(
    #         func=fn,
    #         args=[*args],
    #         keyword=[
    #             ast.keyword(arg=k, value=v)
    #             for k,v in kwargs.items()
    #         ]
    #     )
    # @classmethod
    # def ast_attr(cls, obj, attr):
    #     if isinstance(obj, str):
    #         obj = cls.ast_var(obj)
    #     return ast.Attribute(value=obj, attr=attr, ctx=ast.Load())
    # @staticmethod
    # def ast_np(fun, *args, **kwargs):
    #     return ast.Call(
    #         func=ast.Attribute(value=ast.Name(id='np', ctx=ast.Load()), attr=fun, ctx=ast.Load()),
    #         args=[*args],
    #         keyword=[
    #             ast.keyword(arg=k, value=v)
    #             for k,v in kwargs.items()
    #         ]
    #     )
    # @classmethod
    # def ast_nparray(cls, *args):
    #     return cls.ast_np('array', ast.List(elts=[*args], ctx=ast.Load()))

    def compile(self, mode='numba'):
        self.reset_var()
        expr = eval(
            compile(
                self.get_compile_spec().to_eval_expr(),
                "<expression>", mode='eval'
            ),
            {'np':np}
        )
        if mode=='numba':
            from ...Misc import njit
            expr = njit(expr)
        return expr
    @abc.abstractmethod
    def deriv(self, *which, simplify=True) -> 'Functionlike':
        ...

    def __call__(self, r):
        if isinstance(r, Functionlike):
            return self.compose(r)
        else:
            r = np.asanyarray(r)
            return self.eval(r)

    def __neg__(self):
        return -1*self
    def __mul__(self, other):
        if isinstance(other, (int, float, np.integer, np.floating)):
            other = Scalar(other)
        return Product.construct(self, other)
        # if self.idx is None or other.idx is None
        # return Product(self, other)
    def __rmul__(self, other):
        return self * other
    def __truediv__(self, other):
        # print("t???", self, other, Power(-1).compose(other))
        return self * Power(-1).compose(other)
    def __rtruediv__(self, other):
        # print("r???", self, other)
        return other * Power(-1).compose(self)

    def __add__(self, other):
        return Summation.construct(self, other)
    def __radd__(self, other):
        return Summation.construct(self, other)
    def __sub__(self, other):
        return Summation.construct(self, -other)
    def __rsub__(self, other):
        return Summation.construct(other, -self)

    def __pow__(self, power, modulo=None):
        return Power(power).compose(self)

    def invert(self):
        raise NotImplementedError("{} is not invertible".format(self))

    def __invert__(self):
        return self.invert()

    def copy(self):
        return copy.copy(self)

    def compose(self, other):
        return Composition.construct(self, other)

    @staticmethod
    def is_zero(f:'Functionlike'):
        return isinstance(f, Scalar) and f.scalar == 0
    @staticmethod
    def is_one(f:'Functionlike'):
        return isinstance(f, Scalar) and f.scalar == 1
    @staticmethod
    def is_identity(f:'Functionlike'):
        return isinstance(f, Identity)
    
    @property
    def sort_val(self):
        return self.get_sortval()
    def get_sortval(self):
        return -1

    def simplify(self, iterations=10)->'Functionlike':
        old = new = self
        for r in range(iterations):
            new = old.apply_simplifications()
            if new == old:
                break
            old = new
        return new
    def apply_simplifications(self)->'Functionlike':
        return self

    @classmethod
    def merge_funcs(cls, funcs, reducer, iterations=10):
        for _ in range(iterations):
            unused = set(range(len(funcs)))
            new_funcs = []
            changed = False
            for i1, i2 in itertools.combinations(range(len(funcs)), 2):
                new = reducer(funcs[i1], funcs[i2])
                if new is not None:
                    changed = True
                    new_funcs.extend(new)
                    unused.remove(i1)
                    unused.remove(i2)
                    break
            new_funcs.extend(funcs[i] for i in unused)
            funcs = list(new_funcs)
            if not changed:
                break
        return funcs

    def __hash__(self):
        return id(self)

    def get_children(self):
        return []
    @property
    def children(self):
        return self.get_children()

    @classmethod
    def traverse(cls,
                 root,
                 traversal_order='depth',
                 visit_order='post',
                 node_test=None,
                 max_depth=None,
                 track_index=False
                 ):

        queue = collections.deque()
        if traversal_order == 'depth':
            queue_up = queue.appendleft
        else:
            queue_up = queue.appendleft

        queue_up(root if not track_index else ([], root))

        visit_queue = collections.deque()
        if visit_order == 'post':
            visit = visit_queue.appendleft
        else:
            visit = visit_queue.append

        if node_test is None:
            node_test = lambda n:True

        depth = 0
        while queue and (max_depth is None or depth < max_depth):
            cur = queue.popleft()
            if track_index:
                pos, cur = cur
            for n,c in enumerate(cur.children):
                if node_test(c):
                    if track_index:
                        queue_up((pos+[n], c))
                    else:
                        queue_up(c)
            if track_index:
                visit((pos, cur))
            else:
                visit(cur)
        return visit_queue

    def get_child(self, pos)->'Functionlike':
        raise NotImplementedError("{} is atomic".format(self))
    def replace_child(self, pos, new)->'Functionlike':
        raise NotImplementedError("{} is atomic".format(self))

    def tree_repr(self, sep="", indent=""):
        return "{}()".format(type(self).__name__)

class ElementaryFunction(Functionlike):
    """
    A _univariate_ function (though it can be threadable)
    that has known values and derivatives
    """
    __slots__ = ['idx']

    def __init__(self, *, idx=None):
        self.idx=idx

    @abc.abstractmethod
    def get_deriv(self) -> 'ElementaryFunction':
        ...

    def deriv(self, order=1, simplify=True):  # this can be overloaded but for now this is the easy way to bootstrap
        d = self.get_deriv()
        for _ in range(order - 1):
            d = d.get_deriv()
        if simplify:
            d = d.simplify()
        return d

    # @classmethod
    # def __getitem__(cls, item):  # to support Morse[0](r...) syntax
    #     return cls(idx=item)

    def __repr__(self):
        return "{}[{}]".format(
            type(self),
            self.idx
        )

    def idx_compatible(self, other):
        return (
                isinstance(other, Scalar) or # invariant upon change of indices
                isinstance(self, Scalar) or
                other.idx is None or self.idx is None or other.idx == self.idx
        )

class Variable(ElementaryFunction):
    def __init__(self, name, idx):
        super().__init__(idx=idx)
        self.name = name
    def eval(self, r: np.ndarray) -> 'np.ndarray':
        return r
    def get_compile_spec(self):
        return Abstract.Lambda(Abstract.Name(self.name))(Abstract.Name(self.name))
    def get_deriv(self) -> 'Functionlike':
        return Scalar(1)
    def simplify(self, iterations=10) ->'Functionlike':
        return self
    def __hash__(self):
        return hash((Variable, self.name, self.idx))
    def __eq__(self, other):
        return (
                isinstance(other, Variable)
                and other.name == self.name
                and self.idx_compatible(other)
        )
    def __repr__(self):
        return self.name
    def tree_repr(self, sep="", indent=""):
        return "{}({})".format(type(self).__name__, self.name)

class ElementaryVaradic(ElementaryFunction):
    __slots__ = ['functions', 'idx']
    def __init__(self, *functions: ElementaryFunction, idx=None):
        if idx is None:
            for f in functions:
                if f.idx is not None:
                    idx = f.idx
                    break
        super().__init__(idx=idx)
        self.functions = functions

    sort_key = None
    def get_sortval(self):
        return self.sort_key + sum(f.sort_val for f in self.functions)
    def __hash__(self):
        return hash((type(self), self.functions, self.idx))
    def tree_equivalent(self, other):
        return (
                len(other.functions) == len(self.functions)
                and self.idx_compatible(other)
                and all(f1 == f2 for f1, f2 in zip(self.functions, other.functions))
        )
    def __eq__(self, other):
        return isinstance(other, type(self)) and self.tree_equivalent(other)
    def get_children(self):
        return self.functions

    @classmethod
    @abc.abstractmethod
    def get_repr(cls, fns)->str:
        ...
    def __repr__(self):
        return self.get_repr(self.functions)

    def tree_repr(self, sep="\n", indent=""):
        return "{name}({s}{terms}{s})".format(
            name=type(self).__name__,
            s=sep+indent,
            terms=(","+sep+indent).join(f.tree_repr(sep=sep, indent=indent+" ") for f in self.functions)
        )

    def get_child(self, pos):
        if isinstance(pos, (int, np.integer)):
            return self.functions[pos]
        else:
            fn = self
            for p in pos:
                fn = fn.get_child(p)
        return fn
    def replace_child(self, pos, new)->'ElementaryVaradic':
        if not isinstance(pos, (int, np.integer)):
            if len(pos) == 1:
                pos = pos[0]
            else:
                new = self.get_child(pos[0]).replace_child(pos[1:], new)
                pos = pos[0]

        fn = self.copy()
        fn.functions = fn.functions[:pos] + (new,) + fn.functions[pos+1:] # ez?
        return fn

class ElementarySummation(ElementaryVaradic):
    sort_key = 1000

    def eval(self, r:np.ndarray) ->'np.ndarray':
        return np.sum([f(r) for f in self.functions], axis=0)
    def get_compile_spec(self):
        var = self.get_compile_var()
        np = Abstract.Name('np')
        args = [f.get_compile_spec()(var) for f in self.functions]
        return Abstract.Lambda(var)(reduce(lambda x,y:x+y, args))
    def get_deriv(self) ->'ElementaryFunction':
        fns = self.functions
        return type(self)(*[f.deriv() for f in fns])
    @classmethod
    def get_repr(cls, fns):
        return "+".join(repr(x) for x in fns)

    @classmethod
    def merge_product(cls, f1, f2):
        if isinstance(f1, ElementaryProduct):
            scalars1 = [f.scalar for f in f1.functions if isinstance(f, Scalar)]
            terms1 = [f for f in f1.functions if not isinstance(f, Scalar)]
        else:
            scalars1 = [1]
            terms1 = [f1]

        if isinstance(f2, ElementaryProduct):
            scalars2 = [f.scalar for f in f2.functions if isinstance(f, Scalar)]
            terms2 = [f for f in f2.functions if not isinstance(f, Scalar)]
        else:
            scalars2 = [1]
            terms2 = [f2]

        if len(terms1) == len(terms2) and all(t1 == t2 for t1, t2 in zip(terms1, terms2)):
            return ElementaryProduct(Scalar(sum(scalars1)+sum(scalars2)), *terms1)

    @classmethod
    def reduce_pair(cls, f1, f2)->'Iterable[Functionlike]|bool':
        # implement possible reduction rules for sums
        mp = cls.merge_product(f1, f2)
        return [mp] if mp is not None else mp

    def apply_simplifications(self):
        simp_funs = [f.simplify() for f in sorted(self.functions, key=lambda f:f.sort_val)]
        newfs = [
            x
            for f in simp_funs
            for x in (f.functions if isinstance(f, ElementarySummation) else [f])
            if not self.is_zero(f)
        ]

        newfs = self.merge_funcs(newfs, self.reduce_pair)

        if len(newfs) == 0:
            return Scalar(0)
        elif len(newfs) == 1:
            return newfs[0]
        else:
            return type(self)(*sorted(newfs, key=lambda f: f.sort_val), idx=self.idx)

class ElementaryProduct(ElementaryVaradic):
    sort_key = 100

    def eval(self, r:np.ndarray) ->'np.ndarray':
        return np.prod([f(r) for f in self.functions], axis=0)
    def get_compile_spec(self):
        var = self.get_compile_var()
        np = Abstract.Name('np')
        args = [
            f.get_compile_spec()(var)
            for f in self.functions
        ]
        return Abstract.Lambda(var)(reduce(lambda x,y:x*y, args))
    def get_deriv(self) ->'ElementaryFunction':
        fns = self.functions
        prod = type(self)
        return ElementarySummation(*[
            prod(*[f if j!=i else f.deriv() for j,f in enumerate(fns)])
            for i in range(len(fns))
        ])

    @classmethod
    def reduce_pair(cls, f1, f2) -> 'Iterable[Functionlike]|bool':
        # implement possible reduction rules for products
        if isinstance(f1, Scalar) and isinstance(f2, Scalar):
            return [Scalar(f1.scalar*f2.scalar)]

    def apply_simplifications(self):
        simp_funs = [f.simplify() for f in sorted(self.functions, key=lambda f: f.sort_val)]
        for f in simp_funs:
            if self.is_zero(f):
                return Scalar(0)

        newfs = [
            x
            for f in simp_funs
            for x in (f.functions if isinstance(f, ElementaryProduct) else [f])
            if not self.is_one(f)
        ]

        newfs = self.merge_funcs(newfs, self.reduce_pair)
        
        if len(newfs) == 0:
            return Scalar(1)
        elif len(newfs) == 1:
            return newfs[0]
        else:
            return type(self)(*sorted(newfs, key=lambda f: f.sort_val), idx=self.idx)

    @classmethod
    def get_repr(cls, fns):
        return "*".join("({})".format(r) if isinstance(r, ElementarySummation) else repr(r) for r in fns)

class ElementaryComposition(ElementaryVaradic):
    sort_key = 0

    def eval(self, r:np.ndarray) ->'np.ndarray':
        return reduce(lambda r,f:f(r), reversed(self.functions), r)

    def get_compile_spec(self):
        var = self.get_compile_var()
        return Abstract.Lambda(var)(
            reduce(
                lambda r, f: f.get_compile_spec()(r),
                reversed(self.functions),
                var
            )
        )

    def get_deriv(self) ->'ElementaryFunction':
        rev = list(reversed(self.functions))
        return ElementaryProduct(*[
            type(self)(f.get_deriv(), *rev[:i], idx=self.idx)
                if i > 0 else
            f.get_deriv()

            for i,f in enumerate(rev)
                if not isinstance(f, Variable)
        ])

    @classmethod
    def subs_identities(cls, f1, f2):
        idpos = cls.traverse(f1, track_index=True)
        changed = False
        for pos,node in idpos:
            if cls.is_identity(node):
                changed = True
                f1 = f1.replace_child(pos, f2)
        return changed, f1

    def apply_simplifications(self):
        simp_funs = [f.simplify() for f in self.functions]
        newf_point = 0
        for i,f in enumerate(simp_funs):
            if isinstance(f, Scalar):
                newf_point = i

        newfs = [
            x
            for f in simp_funs[newf_point:]
            for x in (f.functions if isinstance(f, ElementaryComposition) else [f])
            if not self.is_identity(f)
        ]

        # print("="*20)
        # for x in newfs:
        #     print(x.tree_repr())

        new_newfs = []
        curf = None
        for i in range(len(newfs)-1, -1, -1):# count backwards
            if curf is None:
                curf = newfs[i]
            else:
                changed, new = self.subs_identities(newfs[i], curf) # try to replace identities in newfs[i]
                if changed:
                    curf = new
                else:
                    new_newfs.append(curf)
                    curf = newfs[i]
        new_newfs.append(curf)
        newfs = list(reversed(new_newfs))
        # print("-"*10)
        # for x in newfs:
        #     print(x.tree_repr())


        if len(newfs) == 1:
            return newfs[0]
        elif len(newfs) == 0:
            return Identity()
        else:
            return type(self)(*newfs, idx=self.idx)

    @classmethod
    def get_repr(cls, fns):
        fmt_string = "{}({})"
        reprs = [repr(f) for f in fns]
        return reduce(
            lambda a, b:
            fmt_string.format(a, b)
            if "{}" not in a else
            a.replace("{}", "{0}").format(b if all(op not in b for op in ['+', "-", "*"]) else "({})".format(b)),
            reprs
        )

class MultivariateFunction(Functionlike):
    """
    A multivariate function composed from elementary functions.
    This is a
    """

    def __init__(self, *functions:'ElementaryFunction|MultivariateFunction', indices=None):
        self.functions = functions
        self._inds = indices

    @classmethod
    def _reduce_indices(cls, terms):
        return {t.idx for t in terms if isinstance(t, ElementaryFunction) and t.idx is not None}
    @classmethod
    def construct_varivariate(cls, univariate, multivariate, terms, indices=None):
        il = cls._reduce_indices(terms)
        return (
            univariate(*terms, idx=indices[0] if indices is not None else indices)
            if len(il) < 2 and all(isinstance(t, ElementaryFunction) for t in terms) else
            multivariate(*terms, indices=indices)
        )

    @property
    def indices(self):
        if self._inds is None:
            terms = [
                [f.idx] if isinstance(f, ElementaryFunction) else f.indices for f in self.functions
                if not isinstance(f, ElementaryFunction) or f.idx is not None
            ]
            self._inds = np.unique(np.concatenate(terms)) if len(terms) > 0 else np.array([], dtype=int)
        return self._inds

    @property
    def ndim(self):
        return len(self.indices)

    @abc.abstractmethod
    def get_deriv(self, *counts) -> 'Functionlike':
        ...

    def deriv(self, *which, order=1, ndim=None, simplify=True) -> 'Functionlike':
        if len(which) == 0:
            ndim = self.ndim if ndim is None else ndim
            res = np.full((ndim,) * order, None)
            # partitions = SymmetricGroupGenerator(ndim).get_terms(order)
            for which in itertools.combinations_with_replacement(range(ndim), r=order):
                count_map = {k: v for k, v in zip(*np.unique(which, return_counts=True))}
                fun = self.get_deriv(*(count_map.get(k, 0) for k in range(self.ndim)))
                for p in itertools.permutations(which):
                    res[p] = fun
            deriv = TensorFunction(res, symmetric=True, indices=self.indices)
        else:
            count_map = {k:v for k,v in zip(*np.unique(which, return_counts=True))}
            deriv = self.get_deriv(*(count_map.get(k, 0) for k in range(self.ndim)))

        if simplify:
            deriv = deriv.simplify()
        return deriv

    sort_key = None
    def get_sortval(self):
        return self.sort_key + sum(f.sort_val for f in self.functions)

    def apply_simplifications(self) ->'Functionlike':
        return type(self)(
            *(f.simplify() for f in self.functions),
            indices=self.indices
        )

    def __hash__(self):
        return hash((type(self), self.functions, self.indices))
    def tree_equivalent(self, other):
        return (
                len(other.functions) == len(self.functions)
                and (self.indices == other.indices).all()
                and all(f1 == f2 for f1, f2 in zip(self.functions, other.functions))
        )

    def __eq__(self, other):
        return isinstance(other, type(self)) and self.tree_equivalent(other)

    def get_children(self):
        return self.functions

    # @classmethod
    # @abc.abstractmethod
    # def get_repr(cls, fns) -> str:
    #     ...
    # def __repr__(self):
    #     return self.get_repr(self.functions)

    def tree_repr(self, sep="\n", indent=""):
        return "{name}({s}{terms}{s})".format(
            name=type(self).__name__,
            s=sep + indent,
            terms=("," + sep + indent).join(f.tree_repr(sep=sep, indent=indent + " ") for f in self.functions)
        )

    def get_child(self, pos):
        if isinstance(pos, (int, np.integer)):
            return self.functions[pos]
        else:
            fn = self
            for p in pos:
                fn = fn.get_child(p)
        return fn
    def replace_child(self, pos, new) -> 'ElementaryVaradic':
        if not isinstance(pos, (int, np.integer)):
            if len(pos) == 1:
                pos = pos[0]
            else:
                new = self.get_child(pos[0]).replace_child(pos[1:], new)
                pos = pos[0]

        fn = self.copy()
        fn.functions = fn.functions[:pos] + (new,) + fn.functions[pos + 1:]  # ez?
        return fn

class TensorFunction(MultivariateFunction):
    """
    A tensor of functions
    """
    def __init__(self, functions:np.ndarray, symmetric=True, indices=None):
        functions = np.asanyarray(functions, dtype=object)
        self.symmetric = symmetric
        super().__init__(*functions.flat, indices=indices)
        self.functions = functions
    def _get_res_array(self, v)->np.ndarray:
        if isinstance(v, np.ndarray):
            res = np.empty(self.functions.shape + v.shape, dtype=v.dtype)
        elif isinstance(v, str):
            res = np.empty(self.functions.shape, dtype=object) # someday maybe I'll treat as strings...?
        elif isinstance(v, (int, np.integer)):
            res = np.empty(self.functions.shape, dtype=int)
        elif isinstance(v, (float, np.floating)):
            res = np.empty(self.functions.shape, dtype=float)
        else:
            res = np.empty(self.functions.shape, dtype=object)
        return res
    def apply_function(self, fn, res_builder=None)->np.ndarray:
        flat = self.functions.flat
        if self.symmetric:
            _cache = {}
        if res_builder is None:
            res_builder = self._get_res_array
        res = None
        for f in flat:
            idx = np.unravel_index(flat.index - 1, self.functions.shape)
            if not self.symmetric:
                v = fn(f)
                if res is None:
                    res = res_builder(v)
                res[idx] = v
            else:
                key = tuple(np.sort(idx))
                if key not in _cache:
                    _cache[key] = fn(f)
                if res is None:
                    res = res_builder(_cache[key])
                res[idx] = _cache[key]
        return res
    def eval(self, r:np.ndarray) ->'np.ndarray':
        return self.apply_function(
            lambda f:
                f(r[..., f.idx] if f.idx is not None else r[..., 0])
                    if isinstance(f, ElementaryFunction) else
                f(r)
        )
    def get_compile_spec(self):
        var = self.get_compile_var()
        _, np = Abstract.vars("_", "np")
        spec_calls = [
            f.get_compile_spec()(
                (var[..., f.idx]  if f.idx is not None else var[..., 0])
                    if (isinstance(f, ElementaryFunction)) else
                var
            )

            for f in self.functions.flat
        ]

        return Abstract.Lambda(var)(
            Abstract.Lambda(_)(
                np.reshape(_, Abstract.Tuple(*self.functions.shape) + _.shape[1:])
            )(
                np.concatenate(tuple(
                    Abstract.Lambda(_)(
                        np.reshape(_, Abstract.Tuple(1) + _.shape)
                    )(x)
                    for x in spec_calls
                ),
                    axis=0
                )
            )
        )
    def get_deriv(self, *counts)->'TensorFunction':
        return TensorFunction(self.apply_function(lambda f:f.get_deriv(*counts)), indices=self.indices)
    def get_sortval(self):
        return self.sort_key + sum(f.sort_val for f in self.functions.flat)
    def apply_simplifications(self) ->'Functionlike':
        return type(self)(self.apply_function(lambda f:f.simplify()), indices=self.indices)
    def __repr__(self):
        return "{}({})".format(type(self).__name__, self.functions)

    @classmethod
    def format_repr_array(cls, arr, ilevel=0, brackets="[]", sep=",\n", indent=" "):
        # # non-recursive depth-first walk of the array tree rather than
        # # itertools based or arr.flat based walk since both of those require
        # # explicit tracking of indices
        # bl, br = brackets
        # queue = collections.deque()
        # queue.append(arr)
        # lines = []
        # while queue:
        #     cur = queue.pop()
        #     if isinstance(cur[0], np.ndarray):
        #         queue.extend(cur)
        #         lines.append(bl)
        #         ilevel += 1
        #     else:
        #         lines.append(indent+bl)
        #         lines.append((sep+indent).join(cur))
        #         lines.append(indent+br)
        #         ilevel -= 1
        # return "\n".join(lines)

        ## Meh. Recursion is easier.
        lines = []
        lines.append((ilevel*indent)+brackets[0])
        if isinstance(arr[0], np.ndarray):
            for row in arr:
                lines.append(cls.format_repr_array(row, ilevel=ilevel+1, brackets=brackets, sep=sep, indent=indent))
        else:
            lines.append((ilevel*indent)+sep.join(arr))
        lines.append((ilevel*indent)+brackets[1])
        return "\n".join(lines)

    def tree_equivalent(self, other):
        return (
                len(other.functions) == len(self.functions)
                and (self.indices == other.indices).all()
                and (self.functions == other.functions).all()
        )
    def tree_repr(self, sep="\n", indent=""):
        return "{name}({terms})".format(
            name=type(self).__name__,
            terms=self.format_repr_array(self.apply_function(lambda f:f.tree_repr(sep=sep, indent=indent + " ")))
        )
    def copy(self)->'TensorFunction':
        cp = super().copy() #type: TensorFunction
        cp.functions = cp.functions.copy()
        return cp

    def get_children(self)->'Iterable[Functionlike]':
        return self.functions.flat
    def get_child(self, pos)->'Functionlike':
        if isinstance(pos, (int, np.integer)):
            pos = np.unravel_index(pos, self.functions.shape)
            return self.functions[pos]
        else:
            fn = self
            for p in pos:
                fn = fn.get_child(p)
        return fn
    def replace_child(self, pos, new) -> 'TensorFunction':
        if not isinstance(pos, (int, np.integer)):
            if len(pos) == 1:
                pos = pos[0]
            else:
                new = self.get_child(pos[0]).replace_child(pos[1:], new)
                pos = pos[0]

        fn = self.copy()
        fn.functions[pos] = new
        return fn

class Summation(MultivariateFunction):
    """
    A summation of 1D functions to support testing derivs
    """

    def eval(self, r: np.ndarray) -> 'np.ndarray':
        vals = [f(r[..., f.idx] if f.idx is not None else r[..., 0]) if isinstance(f, ElementaryFunction) else f(r) for f in self.functions]
        return np.sum(vals, axis=0)
    def get_compile_spec(self):
        var = self.get_compile_var()
        np = Abstract.Name('np')
        args = [
            f.get_compile_spec()(
                (var[..., f.idx]  if f.idx is not None else var[..., 0])
                    if (isinstance(f, ElementaryFunction)) else
                var
            )

            for f in self.functions
        ]
        return Abstract.Lambda(var)(reduce(lambda x,y:x+y, args))

    @classmethod
    def construct(cls, *terms, indices=None):
        terms = [Scalar(x) if isinstance(x, (int, float, np.integer, np.floating)) else x for x in terms]
        terms = [
            x
            for t in terms
            for x in (
                t.functions
                    if isinstance(t, (Summation, ElementarySummation))
                else [t]
            )
        ]
        terms = [t for t in terms if not (isinstance(t, Scalar) and t.scalar == 0)]
        if len(terms) == 0:
            return Scalar(0, idx=0)
        return cls.construct_varivariate(ElementarySummation, cls, terms, indices=indices)

    def get_deriv(self, *counts)->'MultivariateFunction':
        needs_der_pos = {i for i,c in enumerate(counts) if c > 0}
        if len(needs_der_pos) == 1:
            idx = list(needs_der_pos)[0]
        else:
            idx = -1
        return type(self)(
            *(
                f.deriv(counts[idx])
                    if isinstance(f, ElementaryFunction) else
                f.get_deriv(*counts)
                for i, f in enumerate(self.functions)

                if not isinstance(f, ElementaryFunction) or
                   len(needs_der_pos) == 1 and (f.idx is None or f.idx in needs_der_pos)
            ),
            indices=self.indices
        )

    sort_key = 1000

    @classmethod
    def merge_product(cls, f1, f2):
        if isinstance(f1, (Product, ElementaryProduct)):
            scalars1 = [f.scalar for f in f1.functions if isinstance(f, Scalar)]
            terms1 = [f for f in f1.functions if not isinstance(f, Scalar)]
        else:
            scalars1 = [1]
            terms1 = [f1]

        if isinstance(f2,  (Product, ElementaryProduct)):
            scalars2 = [f.scalar for f in f2.functions if isinstance(f, Scalar)]
            terms2 = [f for f in f2.functions if not isinstance(f, Scalar)]
        else:
            scalars2 = [1]
            terms2 = [f2]

        if len(terms1) == len(terms2) and all(t1 == t2 for t1, t2 in zip(terms1, terms2)):
            return type(f1)(Scalar(sum(scalars1) + sum(scalars2)), *terms1)

    @classmethod
    def reduce_pair(cls, f1, f2) -> 'Iterable[Functionlike]|bool':
        # implement possible reduction rules for sums
        mp = cls.merge_product(f1, f2)
        return [mp] if mp is not None else mp

    def apply_simplifications(self):
        simp_funs = [f.simplify() for f in sorted(self.functions, key=lambda f: f.sort_val)]
        newfs = [
            x
            for f in simp_funs
            for x in (f.functions if isinstance(f, (ElementarySummation, Summation)) else [f])
            if not self.is_zero(f)
        ]

        # newfs = self.merge_funcs(newfs, self.reduce_pair)

        if len(newfs) == 0:
            return Scalar(0, idx=0)
        elif len(newfs) == 1:
            return newfs[0]
        else:
            return type(self)(*sorted(newfs, key=lambda f: f.sort_val), indices=self.indices)

    def __repr__(self):
        return ElementarySummation.get_repr(self.functions)

class Product(MultivariateFunction):
    """
    A summation of 1D functions to support testing derivs
    """

    def eval(self, r: np.ndarray) -> 'np.ndarray':
        vals = [f(r[..., f.idx] if f.idx is not None else r[..., 0]) if isinstance(f, ElementaryFunction) else f(r) for f in self.functions]
        return np.prod(vals, axis=0)
    def get_compile_spec(self):
        var = self.get_compile_var()
        args = [
            f.get_compile_spec()(
                (var[..., f.idx] if f.idx is not None else var[..., 0])
                    if (isinstance(f, ElementaryFunction)) else
                var
            )

            for f in self.functions
        ]
        return Abstract.Lambda(var)(reduce(lambda x,y:x*y, args))

    @classmethod
    def construct(cls, *terms, indices=None):
        terms = [Scalar(x) if isinstance(x, (int, float, np.integer, np.floating)) else x for x in terms]
        terms = [
            x
            for t in terms
            for x in (
                t.functions
                if isinstance(t, (Product, ElementaryProduct))
                else [t]
            )
        ]
        for t in terms:
            if isinstance(t, Scalar) and t.scalar == 0:
                return Scalar(0, idx=0)
        terms = [t for t in terms if not (isinstance(t, Scalar) and t.scalar == 1)]
        if len(terms) == 0:
            return Scalar(1, idx=0)
        return cls.construct_varivariate(ElementaryProduct, cls, terms, indices=indices)

    def get_1d_deriv(self, idx, amt):
        # basically identical to ElementaryFunction version but we filter
        # against irrelevant indices
        counts = [0]*self.ndim
        counts[idx] = amt
        return Summation(
            *(
                # equivalent of ElementaryFunction version
                type(self)(
                    *(
                        (
                            f.deriv(amt)
                                if isinstance(f, ElementaryFunction) else
                            f.get_deriv(*counts)
                        )
                        if j == i else
                        f

                        for j, f in enumerate(self.functions)
                    ),
                    indices=self.indices
                )

                for i in range(len(self.functions))
                if not isinstance(self.functions[i], ElementaryFunction) or
                   (self.functions[i].idx is None or self.functions[i].idx == idx)
            ),
            indices=self.indices
        )

    def get_deriv(self, *counts):
        needs_der_pos = {i for i, c in enumerate(counts) if c > 0}
        if len(needs_der_pos) == 0:
            return self
        elif len(needs_der_pos) == 1:
            idx = list(needs_der_pos)[0]
            return self.get_1d_deriv(idx, counts[idx])
        else:
            for i,c in enumerate(counts):
                if c > 0: # has to happen at some point
                    new = self.get_1d_deriv(i, c)
                    rest = [0]*self.ndim
                    rest[i+1:] = counts[i+1:]
                    return new.get_deriv(*rest)
            else:
                raise ValueError("shouldn't be here...")

    sort_key = 100

    def apply_simplifications(self):
        simp_funs = [f.simplify() for f in self.functions]
        for f in simp_funs:
            if isinstance(f, Scalar) and f.scalar == 0:
                return Scalar(0, idx=0)

        newfs = [
            f for f in simp_funs
            if not (isinstance(f, Scalar) and f.scalar == 1)
        ]
        if len(newfs) == 0:
            return Scalar(1, idx=0)
        return type(self)(*newfs, indices=self.indices)

    def __repr__(self):
        return ElementaryProduct.get_repr(self.functions)

class Composition(MultivariateFunction):
    """
    A composition of multivariate functions that
    uses the chain rule for derivatives
    """

    def eval(self, r: np.ndarray) -> 'np.ndarray':
        return reduce(lambda r,f: f(r[..., f.idx] if f.idx is not None else r[..., 0]), reversed(self.functions), r)
    def get_compile_spec(self):
        var = self.get_compile_var()
        return Abstract.Lambda(var)(
            reduce(
                lambda r, f: f.get_compile_spec()(
                    (r[..., f.idx] if f.idx is not None else r[..., 0])
                        if (isinstance(f, ElementaryFunction)) else
                    r
                ),
                reversed(self.functions),
                var
            )
        )

    @classmethod
    def construct(cls, *terms, indices=None):
        terms = [Scalar(x) if isinstance(x, (int, float, np.integer, np.floating)) else x for x in terms]
        terms = [
            x
            for t in terms
            for x in (
                t.functions
                    if isinstance(t, (Composition, ElementaryComposition))
                else [t]
            )
        ]
        sel = 0
        # for i,t in enumerate(terms):
        #     if isinstance(t, Scalar):
        #         sel = i
        #         break
        # terms = terms[:sel+1]
        return cls.construct_varivariate(ElementaryComposition, cls, terms, indices=indices)

    def apply_simplifications(self):
        simp_funs = [f.simplify() for f in self.functions]
        # newf_point = 0
        # for i,f in enumerate(simp_funs):
        #     if isinstance(f, Scalar):
        #         newf_point = i
        #         break

        newfs = [f for f in simp_funs if not isinstance(f, Identity)]
        if len(newfs) == 1:
            return newfs[0]
        elif len(newfs) == 0:
            return Identity(idx=0)
        else:
          return type(self)(*newfs, indices=self.indices)

    # def get_deriv(self) ->'ElementaryFunction':
    #     return ElementaryProduct(*[
    #         # this could be evaluated more efficiently, admittedly...
    #         f.get_deriv() *type(self)(*self.functions[:i], idx=self.idx) if i > 0 else f.get_deriv()
    #         for i,f in enumerate(self.functions)
    #     ])

    # def single_deriv(self, f, arg, ndim, coord):
    #     if isinstance(f, ElementaryFunction):
    #         ...
    #     else:
    #         grad = [None]*ndim
    #         for i in range(ndim):
    #             counts = [0] * ndim
    #             counts[i] = 1
    #             grad[i] = f.get_deriv(*counts)
    #
    #         grad = f.get_deriv(c) for c in co

    def get_deriv(self, *counts):
        # do we assume compositions are always of a single arg
        # or do we allow for multiarg compositions???
        #   --> I think by design we have single arg and then we include multiarg later?

        needs_der_pos = {i for i, c in enumerate(counts) if c > 0}
        nonzero = all(
            not isinstance(f, ElementaryFunction) or
                len(needs_der_pos) == 1 and f.idx in needs_der_pos
            for f in self.functions
        )
        if nonzero:
            c = list(needs_der_pos)[0]
            return Product.construct(*(
                (
                    (
                        f.get_deriv(*counts)
                        if not isinstance(f, ElementaryFunction) else
                        f.get_deriv(c)
                    ) * self.construct(*self.functions[:i])
                    if i > 0 else f.get_deriv()
                )
                for i,f in enumerate(self.functions)
            ), indices=self.indices)
        else:
            return Scalar(0, idx=0)

    def __repr__(self):
        return ElementaryComposition.get_repr(self.functions)

class Scalar(ElementaryFunction):
    """
    Broadcasts a constant value
    """
    __slots__ = ['scalar', 'idx']
    def __init__(self, scalar, *, idx=0):
        idx = None
        super().__init__(idx=idx)
        self.scalar = scalar
    def eval(self, r: np.ndarray) -> 'np.ndarray':
        return np.broadcast_to(np.full(1, self.scalar, dtype=r.dtype), r.shape)
    def get_compile_spec(self):
        if isinstance(self.scalar, Functionlike):
            return self.scalar.get_compile_spec()
        else:
            var = self.get_compile_var()
            np = Abstract.Name('np')
            return Abstract.Lambda(var)(
                np.full(var.shape, self.scalar)
            )
    def get_deriv(self) -> 'ElementaryFunction':
        return type(self)(0, idx=self.idx)
    def deriv(self, order=1, *, simplify=True):
        return type(self)(0, idx=self.idx)
    def get_sortval(self):
        return (self.scalar.sort_val if isinstance(self.scalar, Functionlike) else self.scalar)
    def __hash__(self):
        return hash((Scalar, self.scalar, self.idx))
    def __eq__(self, other):
        return (
                isinstance(other, Scalar)
                and self.idx_compatible(other)
                and self.scalar == other.scalar
        )
    def __repr__(self):
        return repr(self.scalar)
    def tree_repr(self, sep="\n", indent=""):
        return "{}({})".format(type(self).__name__, self.scalar)

class Identity(ElementaryFunction):
    """
    Identity function for compositions
    """
    def eval(self, r: np.ndarray) -> 'np.ndarray':
        return r
    def get_compile_spec(self):
        _ = Abstract.Name("_")
        return Abstract.Lambda(_)(_)
    def get_deriv(self) -> 'ElementaryFunction':
        return Scalar(1)
    def deriv(self, order=1, simplify=True):
        return Scalar(1) if order == 1 else Scalar(0)
    def simplify(self, iterations=10) ->'Functionlike':
        return self
    def __hash__(self):
        return hash((Identity, self.idx))
    def __eq__(self, other):
        return (
                isinstance(other, Identity)
                and self.idx_compatible(other)
        )
    def get_sortval(self):
        return 100
    def __repr__(self):
        return "{}"

class Power(ElementaryFunction):
    __slots__ = ['power', 'idx']
    def __init__(self, power, *, idx=None):
        super().__init__(idx=idx)
        self.power = power
    def eval(self, r:np.ndarray) ->'np.ndarray':
        return r**self.power
    def get_compile_spec(self):
        var = self.get_compile_var()
        power = self.power.get_compile_spec()(var) if isinstance(self.power, Functionlike) else self.power
        return Abstract.Lambda(var)(var**power)
    def get_deriv(self) -> 'ElementaryFunction':
        return self.power * type(self)(self.power-1, idx=self.idx)
    def __hash__(self):
        return hash((Power, self.power, self.idx))
    def __eq__(self, other):
        return (
                isinstance(other, Power)
                and self.idx_compatible(other)
                and self.power == other.power
        )
    sort_key = 20
    def get_sortval(self):
        return self.sort_key + (self.power.sort_val if isinstance(self.power, Functionlike) else self.power)
    def apply_simplifications(self) ->'Functionlike':
        power = self.power
        if isinstance(power, Scalar):
            power = power.scalar
        if isinstance(power, (int, np.integer, float, np.floating)):
            if power == 1:
                return Identity()
            elif power == 0:
                return Scalar(1)
        return self
    def __repr__(self):
        return "{{}}^{}".format(self.power)
    def tree_repr(self, sep="\n", indent=""):
        return "{}({})".format(type(self).__name__, self.power.tree_repr(sep=sep, indent=indent+" ") if isinstance(self.power, Functionlike) else self.power)

class Exponent(ElementaryFunction):
    __slots__ = ['base', 'idx']
    def __init__(self, base, *, idx=None):
        super().__init__(idx=idx)
        self.base = base
    def eval(self, r:np.ndarray) ->'np.ndarray':
        return self.base**r
    def get_compile_spec(self):
        var = self.get_compile_var()
        base = self.base.get_compile_spec()(var) if isinstance(self.base, Functionlike) else self.base
        return Abstract.Lambda(var)(base**var)
    def get_deriv(self) -> 'ElementaryFunction':
        return np.log(self.base) * self
    def __hash__(self):
        return hash((Exponent, self.base, self.idx))
    def __eq__(self, other):
        return (
                isinstance(other, Exponent)
                and self.idx_compatible(other)
                and self.base == other.base
        )
    sort_key = 50
    def get_sortval(self):
        return self.sort_key + (self.base.sort_val if isinstance(self.base, Functionlike) else self.base)
    def __repr__(self):
        return "{}^{{}}".format(self.base)
    def tree_repr(self, sep="\n", indent=""):
        return "{}({})".format(type(self).__name__, self.base.tree_repr(sep=sep, indent=indent+" ") if isinstance(self.base, Functionlike) else self.base)
class Exp(Exponent):
    __slots__ = ['idx']
    def __init__(self, *, idx=None):
        super().__init__(np.e, idx=idx)
    def eval(self, r:np.ndarray) ->'np.ndarray':
        return np.exp(r)
    def get_compile_spec(self):
        return Abstract.Name('np').exp
    def get_deriv(self) -> 'ElementaryFunction':
        return self
    def __repr__(self):
        return "exp{}"
    def tree_repr(self, sep="\n", indent=""):
        return "{}()".format(type(self).__name__)

class Logarithm(ElementaryFunction):
    __slots__ = ['base', 'idx']
    def __init__(self, base, *, idx=None):
        super().__init__(idx=idx)
        self.base = base
    def eval(self, r:np.ndarray) ->'np.ndarray':
        return np.log(r)/np.log(self.base)
    def get_deriv(self) -> 'ElementaryFunction':
        return 1/(np.log(self.base)*self)
    def get_compile_spec(self):
        var = self.get_compile_var()
        np = Abstract.Name('np')
        base = self.base.get_compile_spec()(var) if isinstance(self.base, Functionlike) else self.base
        return Abstract.Lambda(var)(
            np.log(var) / np.log(base)
            )
    def __hash__(self):
        return hash((Logarithm, self.base, self.idx))
    def __eq__(self, other):
        return (
                isinstance(other, Logarithm)
                and self.idx_compatible(other)
                and self.base == other.base
        )
    sort_key = 60
    def get_sortval(self):
        return self.sort_key + (self.base.sort_val if isinstance(self.base, Functionlike) else self.base)
    def __repr__(self):
        return "log_{}{{}}".format(self.base)
    def tree_repr(self, sep="\n", indent=""):
        return "{}({})".format(type(self).__name__, self.base.tree_repr(sep=sep, indent=indent+" ") if isinstance(self.base, Functionlike) else self.base)
class Ln(Logarithm):
    __slots__ = ['idx']
    def __init__(self, *, idx=None):
        super().__init__(np.e, idx=idx)
    def eval(self, r:np.ndarray) ->'np.ndarray':
        return np.log(r)
    def get_compile_spec(self):
        return Abstract.Name('np').log
    def get_deriv(self) -> 'ElementaryFunction':
        return 1/self
    def __repr__(self):
        return "ln{}"
    def tree_repr(self, sep="\n", indent=""):
        return "{}()".format(type(self).__name__)

class Sin(ElementaryFunction):
    def eval(self, r:np.ndarray) ->'np.ndarray':
        return np.sin(r)
    def get_deriv(self) -> 'ElementaryFunction':
        return Cos()
    def get_compile_spec(self):
        return Abstract.Name('np').sin
    def __hash__(self):
        return hash((Sin, self.idx))
    def __eq__(self, other):
        return (
                isinstance(other, Sin)
                and self.idx_compatible(other)
        )
    def __repr__(self):
        return "sin{}"
class Cos(ElementaryFunction):
    def eval(self, r:np.ndarray) ->'np.ndarray':
        return np.cos(r)
    def get_compile_spec(self):
        return Abstract.Name('np').cos
    def get_deriv(self) -> 'ElementaryFunction':
        return -Sin()
    def __hash__(self):
        return hash((Cos, self.idx))
    def __eq__(self, other):
        return (
                isinstance(other, Cos)
                and self.idx_compatible(other)
        )
    def __repr__(self):
        return "cos{}"

class CompoundFunction(ElementaryFunction):
    def __init__(self, *, idx=None):
        super().__init__(idx=idx)
        self._expr = None
    @abc.abstractmethod
    def get_expression(self) -> 'ElementaryFunction':
        ...
    @property
    def expression(self):
        if self._expr is None:
            self._expr = self.get_expression()
        return self._expr
    def __eq__(self, other):
        return self.expression == other
    def get_deriv(self) -> 'ElementaryFunction':
        return self.expression.get_deriv()
    def get_compile_spec(self):
        return self.expression.get_compile_spec()
    def get_sortval(self):
        return self.expression.sort_val
class Morse(CompoundFunction):
    def __init__(self, *, de=1, a=1, re=0, idx=None):
        self.de = de
        self.a = a
        self.re = re
        super().__init__(idx=idx)
    def get_expression(self):
        de = self.de if isinstance(self.de, Scalar) else Scalar(self.de)
        a = self.a if isinstance(self.a, Scalar) else Scalar(self.a)
        re = self.re if isinstance(self.re, Scalar) else Scalar(self.re)
        r = Identity(idx=self.idx)
        return de*(Scalar(1)-Exp()(-a*(r-re)))**2
    def eval(self, r: np.ndarray) -> 'np.ndarray':
        return self.de*(1-np.exp(-self.a*(r-self.re )))**2
    def get_deriv(self):
        return MorseDeriv(1, de=self.de, a=self.a, re=self.re, idx=self.idx)
    def __repr__(self):
        return "morse(de={},a={},re={})".format(self.de, self.a, self.re)
    def tree_repr(self, sep="\n", indent=""):
        return "{}(de={},a={},re={})".format(type(self).__name__, self.de, self.a, self.re)
class MorseDeriv(CompoundFunction):
    def __init__(self, order, *, de=1, a=1, re=0, idx=None):
        self.order = order
        self.de = de
        self.a = a
        self.re = re
        super().__init__(idx=idx)
    def eval(self, r: np.ndarray) -> 'np.ndarray':
        r = np.asanyarray(r)
        n = self.order
        de = self.de
        a = self.a
        re = self.re
        return ((-1)**(n + 1) * 2*a**n * de) * np.exp(-2*a*(r-re))*(np.exp(a*(r-re)) - (2 ** (n - 1)))
    def get_expression(self) -> 'ElementaryFunction':
        n = self.order
        de = self.de if isinstance(self.de, Scalar) else Scalar(self.de)
        a = self.a if isinstance(self.a, Scalar) else Scalar(self.a)
        re = self.re if isinstance(self.re, Scalar) else Scalar(self.re)
        r = Identity(idx=self.idx)
        return ((-1) ** (n + 1) * 2 * a ** n * de) * Exp()(-2 * a * (r-re)) * (Exp()(a * (r-re)) - (2 ** (n - 1)))
    def get_deriv(self) -> 'MorseDeriv':
        return type(self)(
            self.order+1,
            de=self.de,
            a=self.a,
            re=self.re,
            idx=self.idx
        )
    def __repr__(self):
        return "morse[{}](de={},a={},re={})".format(self.order, self.de, self.a, self.re)
    def tree_repr(self, sep="\n", indent=""):
        return "{}(order={},de={},a={},re={})".format(type(self).__name__, self.order, self.de, self.a, self.re)

class Symbols:

    def __init__(self, *vars):
        if len(vars) == 1 and isinstance(vars[0], str):
            vars = vars[0]
            if " " in vars:
                vars = vars.split()
        self.vars = [Variable(v, i) if not isinstance(v, Variable) else v for i, v in enumerate(vars)]
        self.__dict__.update({v.name: v for v in self.vars})
    @classmethod
    def scalar(cls, v):
        return Scalar(v)

    @classmethod
    def log(cls, v, base=None):
        return Exp()(v) if base is None else Logarithm(base)(v)

    @classmethod
    def exp(cls, v, base=None):
        return Exp()(v) if base is None else Exponent(base)(v)

    @classmethod
    def cos(cls, x):
        return Cos()(x)

    @classmethod
    def sin(cls, x):
        return Sin()(x)

    @classmethod
    def morse(cls, r, de=1, a=1, re=0):
        return Morse(de=de, a=a, re=re)(r)


class SymPyFunction:
    """
    A function suitable for symbolic manipulation
    with derivatives and evlauation
    """

    @classmethod
    def get_sympy(cls):
        import sympy
        return sympy
    @property
    def sympy(self):
        return self.get_sympy()

    def __init__(self, expr, vars=None):
        if isinstance(expr, str):
            expr = self.sympy.parsing.parse_expr(expr)
        self.expr = expr
        if vars is None:
            vars = self.expr.free_symbols
        self.vars = self.sort_vars(vars)
        self._compiled = None

    def _varkey(self, v):
        if isinstance(v, self.sympy.Symbol):
            v = v.name
        bits = v.split('[', 2)
        if len(bits) > 1:
            v = int(bits[1].split("]")[0])
        return v
    def sort_vars(self, vars):
        return list(sorted(vars, key=lambda v:self._varkey(v)))
    def merge_vars(self, v1, v2):
        v = set(v1)
        v.update(v2)
        return self.sort_vars(v)

    def compile(self):
        if self._compiled is None:
            self._compiled = self.sympy.lambdify(self.vars, self.expr)
        return self._compiled

    def eval(self, r: np.ndarray) -> 'np.ndarray':
        if len(self.vars) == 1:
            if r.ndim > 1 and r.shape[-1] == 1:
                return self.compile()(
                    np.moveaxis(r, -1, 0)[0]
                )
            else:
                return self.compile()(r)
        else:
            return self.compile()(*np.moveaxis(r, -1, 0))

    def deriv(self, *which, order=1):
        cls = type(self)
        if len(self.vars) == 1:
            if len(which) > 0:
                order = which[0]
            expr = self.expr
            for _ in range(order):
                expr = self.sympy.diff(expr, self.vars)
            deriv = cls(expr, vars=self.vars)
        else:
            ndim = len(self.vars)
            if len(which) == 0:
                res = np.full((ndim,) * order, None)
                # partitions = SymmetricGroupGenerator(ndim).get_terms(order)
                for which in itertools.combinations_with_replacement(range(ndim), r=order):
                    count_map = {k: v for k, v in zip(*np.unique(which, return_counts=True))}
                    expr = self.expr
                    for k in range(ndim):
                        for _ in range(count_map.get(k, 0)):
                            expr = self.sympy.diff(expr, self.vars[k])
                    fn = cls(expr, vars=self.vars)
                    for p in itertools.permutations(which):
                        res[p] = expr
                deriv = SymPyArrayFunction(res, symmetric=True)
            else:
                count_map = {k: v for k, v in zip(*np.unique(which, return_counts=True))}
                expr = self.expr
                for k in range(ndim):
                    for _ in range(count_map.get(k, 0)):
                        expr = self.sympy.diff(expr, self.vars[k])
                deriv = cls(expr, vars=self.vars)

        # if simplify:
        #     deriv = deriv.simplify()
        return deriv

    def __call__(self, r):
        if isinstance(r, SymPyFunction):
            return self.compose(r)
        else:
            r = np.asanyarray(r)
            return self.eval(r)

    def __neg__(self):
        return -1 * self

    def __mul__(self, other):
        if isinstance(other, SymPyFunction):
            ovars = other.vars
            other = other.expr
        else:
            ovars = []
        return type(self)(self.expr * other, vars=self.merge_vars(self.vars, ovars))
    def __rmul__(self, other):
        return self * other
    def __truediv__(self, other):
        if isinstance(other, SymPyFunction):
            ovars = other.vars
            other = other.expr
        else:
            ovars = []
        return type(self)(self.expr / other, vars=self.merge_vars(self.vars, ovars))
    def __rtruediv__(self, other):
        if isinstance(other, SymPyFunction):
            ovars = other.vars
            other = other.expr
        else:
            ovars = []
        return type(self)(other / self.expr, vars=self.merge_vars(self.vars, ovars))

    def __add__(self, other):
        if isinstance(other, SymPyFunction):
            ovars = other.vars
            other = other.expr
        else:
            ovars = []
        return type(self)(other + self.expr, vars=self.merge_vars(self.vars, ovars))
    def __radd__(self, other):
        if isinstance(other, SymPyFunction):
            ovars = other.vars
            other = other.expr
        else:
            ovars = []
        return type(self)(other + self.expr, vars=self.merge_vars(self.vars, ovars))

    def __sub__(self, other):
        if isinstance(other, SymPyFunction):
            ovars = other.vars
            other = other.expr
        else:
            ovars = []
        return type(self)(self.expr - other, vars=self.merge_vars(self.vars, ovars))
    def __rsub__(self, other):
        if isinstance(other, SymPyFunction):
            ovars = other.vars
            other = other.expr
        else:
            ovars = []
        return type(self)(other - self.expr, vars=self.merge_vars(ovars, self.vars))

    def __pow__(self, power, modulo=None):
        if isinstance(power, SymPyFunction):
            ovars = power.vars
            power = power.expr
        else:
            ovars = []
        return type(self)(self.expr**power, vars=self.merge_vars(self.vars, ovars))

    def invert(self):
        return type(self)(self.expr.invert())
    def __invert__(self):
        return self.invert()
    def copy(self):
        return copy.copy(self)
    def compose(self, other):
        if isinstance(other, SymPyFunction):
            ovars = other.vars
            other = other.expr
        else:
            ovars = []
        return type(self)(
            self.expr(other),
            vars=self.merge_vars(self.vars, ovars)
        )

    @classmethod
    def symbols(cls, *syms):
        sympy = cls.get_sympy()
        return [cls(sympy.Symbol("{}[{}]".format(s, n))) for n,s in enumerate(syms)]
    @classmethod
    def exp(cls, fn):
        return cls(cls.get_sympy().exp(fn.expr), vars=fn.vars)
    @classmethod
    def morse(cls, var, de=10, a=1, re=0):
        if not isinstance(var, SymPyFunction):
            var = cls(var)
        return de*(1-cls.exp(-a*(var-re)))**2

    def __repr__(self):
        return repr(self.expr)

class SymPyArrayFunction:
    def __init__(self, expr_array, symmetric=False):
        self.functions = expr_array
        self.symmetric = symmetric
        self.functions = self.apply_function(lambda f:SymPyFunction(f) if not isinstance(f, (SymPyFunction, SymPyArrayFunction)) else f)
    def _get_res_array(self, v)->np.ndarray:
        if isinstance(v, np.ndarray):
            res = np.empty(self.functions.shape + v.shape, dtype=v.dtype)
        elif isinstance(v, str):
            res = np.empty(self.functions.shape, dtype=object) # someday maybe I'll treat as strings...?
        elif isinstance(v, (int, np.integer)):
            res = np.empty(self.functions.shape, dtype=int)
        elif isinstance(v, (float, np.floating)):
            res = np.empty(self.functions.shape, dtype=float)
        else:
            res = np.empty(self.functions.shape, dtype=object)
        return res
    def apply_function(self, fn, res_builder=None)->np.ndarray:
        flat = self.functions.flat
        if self.symmetric:
            _cache = {}
        if res_builder is None:
            res_builder = self._get_res_array
        res = None
        for f in flat:
            idx = np.unravel_index(flat.index - 1, self.functions.shape)
            if not self.symmetric:
                v = fn(f)
                if res is None:
                    res = res_builder(v)
                res[idx] = v
            else:
                key = tuple(np.sort(idx))
                if key not in _cache:
                    _cache[key] = fn(f)
                if res is None:
                    res = res_builder(_cache[key])
                res[idx] = _cache[key]
        return res
    def eval(self, r:np.ndarray) ->'np.ndarray':
        return self.apply_function(lambda f:f(r))

    def __call__(self, r):
        if isinstance(r, SymPyFunction):
            return self.compose(r)
        else:
            r = np.asanyarray(r)
            return self.eval(r)

    def __neg__(self):
        return -1 * self

    def __mul__(self, other):
        if isinstance(other, SymPyArrayFunction):
            other = other.functions
        elif isinstance(other, SymPyFunction):
            other = other.expr
        return type(self)(self.functions * other)
    def __rmul__(self, other):
        return self * other

    def __truediv__(self, other):
        if isinstance(other, SymPyArrayFunction):
            other = other.functions
        elif isinstance(other, SymPyFunction):
            other = other.expr
        return type(self)(self.functions / other)
    def __rtruediv__(self, other):
        if isinstance(other, SymPyArrayFunction):
            other = other.functions
        elif isinstance(other, SymPyFunction):
            other = other.expr
        return type(self)(other / self.functions)

    def __add__(self, other):
        if isinstance(other, SymPyArrayFunction):
            other = other.functions
        elif isinstance(other, SymPyFunction):
            other = other.expr
        return type(self)(self.functions + other)
    def __radd__(self, other):
        if isinstance(other, SymPyArrayFunction):
            other = other.functions
        elif isinstance(other, SymPyFunction):
            other = other.expr
        return type(self)(self.functions + other)

    def __sub__(self, other):
        if isinstance(other, SymPyArrayFunction):
            other = other.functions
        elif isinstance(other, SymPyFunction):
            other = other.expr
        return type(self)(self.functions - other)
    def __rsub__(self, other):
        if isinstance(other, SymPyArrayFunction):
            other = other.functions
        elif isinstance(other, SymPyFunction):
            other = other.expr
        return type(self)(other - self.functions)

    def __pow__(self, power, modulo=None):
        if isinstance(power, SymPyArrayFunction):
            power = power.functions
        elif isinstance(power, SymPyFunction):
            power = power.expr
        return type(self)(self.functions ** power)

    def invert(self):
        raise NotImplementedError("...")
        return type(self)(self.apply_function(lambda f:f.invert()))
    def __invert__(self):
        return self.invert()

    def copy(self):
        return copy.copy(self)

    def compose(self, other):
        if isinstance(other, SymPyArrayFunction):
            other = other.functions
        elif isinstance(other, SymPyFunction):
            other = other.expr
        return type(self)(self.apply_function(lambda f:f(other)))
    def __repr__(self):
        return "{}({})".format(type(self).__name__, self.functions)


from .AffineTransform import AffineTransform
import numpy as np

######################################################################################################
##
##                                   ScalingTransform Class
##
######################################################################################################

__all__ = [
    "ScalingTransform"
]
__reload_hook__ = [".AffineTransform"]

class ScalingTransform(AffineTransform):
    """A simple ScalingTransform from the basic AffineTransformation class

    """

    def __init__(self, scalings):
        """creates an AffineTransform

        :param scalings:
        :type scalings:
        """
        super().__init__(np.diag(scalings), shift=None)
